"""Middleware for page states.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from djblets.pagestate.state import PageState
from djblets.util.http import encode_etag

if TYPE_CHECKING:
    from typing import Callable

    from django.http import HttpRequest, HttpResponseBase


def PageStateMiddleware(
    get_response: Callable[[HttpRequest], HttpResponseBase],
) -> Callable[[HttpRequest], HttpResponseBase]:
    """Middleware for updating ETags with those from page state injections.

    Version Added:
        6.0

    Args:
        get_response (callable):
            The function for getting a response from a request.

    Returns:
        callable:
        The middleware callable for processing the request.
    """
    def _middleware(
        request: HttpRequest,
    ) -> HttpResponseBase:
        """Process the HTTP response.

        If the response contains an ETag, any injected ETag data in the
        page state will be merged into it.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

        Returns:
            django.http.HttpResponseBase:
            The resulting HTTP response.
        """
        response = get_response(request)

        if ((old_etag := response.headers.get('ETag')) and
            (page_state := PageState.for_request(request,
                                                 only_if_exists=True)) and
            (injected_etag := page_state.get_etag())):
            # Merge the existing ETag and new one together.
            new_etag = encode_etag(''.join([
                old_etag,
                injected_etag,
            ]).encode('utf-8'))

            response['ETag'] = new_etag

        return response

    return _middleware



"""Common symbols useful for function signatures.

Version Changed:
    6.0:
    This module now just forwards to :py:mod:`typelets.symbols`. Imports to
    this module will be deprecated in a future release.

Version Added:
    3.3
"""

from __future__ import annotations

from typelets.symbols import UNSET, UnsetSymbol, Unsettable


__all__ = [
    'UNSET',
    'UnsetSymbol',
    'Unsettable',
]



"""Basic version and package information."""

from __future__ import annotations

from djblets._version import (
    VERSION,
    __version__,
    __version_info__,
    get_package_version,
    get_version_string,
    is_release,
)


__all__ = [
    'VERSION',
    '__version__',
    '__version_info__',
    'get_package_version',
    'get_version_string',
    'is_release',
]



"""Version information for Djblets.

Version Added:
    6.0:
    Moved these from :file:`djblets/__init__.py`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import TypeAlias

    #: Type for version tuples.
    #:
    #: Version Added:
    #:     6.0
    _Version: TypeAlias = tuple[int, int, int, str, int, bool]


# The version of Djblets.
#
# This is in the format of:
#
#   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#
VERSION: _Version = (6, 1, 0, 'final', 0, True)


def get_version_string() -> str:
    """Return the Djblets version as a human-readable string.

    Returns:
        str:
        The Djblets version.
    """
    major, minor, micro, tag, release_num, is_release = VERSION
    version = f'{major}.{minor}'

    if micro:
        version += f'.{micro}'

    if tag != 'final':
        if VERSION[3] == 'rc':
            version += f' RC{release_num}'
        else:
            version += f' {tag} {release_num}'

    if not is_release:
        version += ' (dev)'

    return version


def get_package_version() -> str:
    """Return the Djblets version as a Python package version string.

    Returns:
        str:
        The Djblets package version.
    """
    major, minor, micro, tag, release_num = VERSION[:5]

    version = f'{major}.{minor}'

    if micro:
        version += f'.{micro}'

    if tag != 'final':
        if tag == 'alpha':
            tag = 'a'
        elif tag == 'beta':
            tag = 'b'

        version += f'{tag}{release_num}'

    return version


def is_release() -> bool:
    """Return whether this is a released version of Djblets.

    Returns:
        bool:
        ``True`` if this is an official release.
    """
    return VERSION[5]


#: An alias for the version information from :py:data:`VERSION`.
#:
#: This does not include the last entry in the tuple (the release state).
__version_info__ = VERSION[:-1]


#: An alias for the package version string.
__version__ = get_package_version()



"""Miscellaneous, useful decorators."""

from __future__ import annotations

import inspect
from functools import update_wrapper, wraps
from typing import TYPE_CHECKING, overload

from django import template
from djblets.deprecation import RemovedInDjblets80Warning
from django.template.library import parse_bits
from django.utils.functional import cached_property as django_cached_property

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping, Sequence
    from typing import Any, Concatenate, ParamSpec, TypeAlias, TypeVar

    from django.template import Context, Node, NodeList
    from django.template.base import FilterExpression, Parser, Token

    _P = ParamSpec('_P')
    _R = TypeVar('_R')
    _TagFunc: TypeAlias = Callable[Concatenate[Context, NodeList, _P], _R]
    _TagCompiler: TypeAlias = Callable[[Parser, Token], Node]


# The decorator decorator.  This is copyright unknown, verbatim from
# http://wiki.python.org/moin/PythonDecoratorLibrary
def simple_decorator(decorator):
    """This decorator can be used to turn simple functions
       into well-behaved decorators, so long as the decorators
       are fairly simple. If a decorator expects a function and
       returns a function (no descriptors), and if it doesn't
       modify function attributes or docstring, then it is
       eligible to use this. Simply apply @simple_decorator to
       your decorator and it will automatically preserve the
       docstring and function attributes of functions to which
       it is applied."""
    def new_decorator(f):
        g = decorator(f)
        update_wrapper(g, f)
        return g
    # Now a few lines needed to make simple_decorator itself
    # be a well-behaved decorator.
    update_wrapper(new_decorator, decorator)
    return new_decorator


def augment_method_from(klass):
    """Augments a class's method with new decorators or documentation.

    This is useful when a class needs to add new decorators or new
    documentation to a parent class's method, without changing the behavior
    or burying the existing decorators.

    The methods using this decorator can provide code to run at the end of
    the parent function. Usually, though, it will just have an empty body
    of ``pass``.
    """
    def _dec(func):
        @wraps(func)
        def _call(*args, **kwargs):
            try:
                f = augmented_func(*args, **kwargs)
            finally:
                func(*args, **kwargs)

            return f

        augmented_func = getattr(klass, func.__name__)

        # Go beyond what @wraps does. Choose a suitable docstring depending
        # on which is set, and combine dictionaries with `func` taking
        # priority.
        _call.__doc__ = func.__doc__ or augmented_func.__doc__
        _call.__dict__.update(augmented_func.__dict__)
        _call.__dict__.update(func.__dict__)

        real_func = _call.__dict__.get('_augmented_func', augmented_func)
        _call.__dict__['_augmented_func'] = real_func

        return _call

    return _dec


class _BlockTagNode(template.Node):
    """Node used to render block tags.

    This is an internal class returned by :py:func:`blocktag`, used to
    interface with a provided function for processing a block of nodes.
    It is not subject to any API stability guarantees.

    Version Added:
        6.0:
        This was moved out from inside of :py:func:`blocktag`.
    """

    ######################
    # Instance variables #
    ######################

    #: The list of nodes inside the block to process.
    node_list: NodeList

    #: Whether to resolve variables passed to the template tag.
    resolve_vars: bool

    #: The positional argument bits passed to the template tag.
    tag_args: Sequence[FilterExpression]

    #: The function to call to handle the block's contents.
    tag_func: _TagFunc

    #: The positional keyword bits passed to the template tag.
    tag_kwargs: Mapping[str, FilterExpression]

    def __init__(
        self,
        *,
        tag_func: _TagFunc,
        nodelist: NodeList,
        tag_args: Sequence[FilterExpression],
        tag_kwargs: Mapping[str, FilterExpression],
        resolve_vars: bool,
    ) -> None:
        """Initialize the node.

        Args:
            tag_func (callable):
                The function to call to handle the block's contents.

            nodelist (django.template.NodeList):
                The list of nodes inside the block to process.

            tag_args (list of str):
                The positional argument bits passed to the template tag.

            tag_kwargs (list of str):
                The positional keyword bits passed to the template tag.

            resolve_vars (bool, optional):
                Whether to resolve variables passed to the template tag.
        """
        self.tag_func = tag_func
        self.nodelist = nodelist
        self.tag_args = tag_args
        self.tag_kwargs = tag_kwargs
        self.resolve_vars = resolve_vars

    def render(
        self,
        context: Context,
    ) -> str:
        """Render the contents of the block.

        This will resolve any variables, if requested, and pass everything
        to the wrapped function.

        Args:
            context (django.template.Context):
                The current template context.

        Returns:
            str:
            The resulting rendered content.
        """
        tag_args = self.tag_args
        tag_kwargs = self.tag_kwargs

        if self.resolve_vars:
            tag_args = [
                var.resolve(context)
                for var in tag_args
            ]

            tag_kwargs = {
                name: var.resolve(context)
                for name, var in tag_kwargs.items()
            }
        else:
            tag_args = [
                var.token
                for var in tag_args
            ]

            tag_kwargs = {
                name: var.token
                for name, var in tag_kwargs.items()
            }

        return self.tag_func(context, self.nodelist, *tag_args, **tag_kwargs)


@overload
def blocktag(
    func: _TagFunc[_P, _R],
) -> _TagCompiler:
    ...


@overload
def blocktag(
    *,
    end_prefix: str = 'end',
    resolve_vars: bool = True,
) -> Callable[[_TagFunc[_P, _R]], _TagCompiler]:
    ...


def blocktag(
    func: (_TagFunc[_P, _R] | None) = None,
    *args,
    end_prefix: str = 'end',
    name: (str | None) = None,
    resolve_vars: bool = True,
    **kwargs,
) -> Any:
    """Create a block template tag with beginning/end tags.

    This does all the hard work of creating a template tag that can
    parse the arguments passed in and then parse all nodes between a
    beginning and end tag (such as ``myblock``/``endmyblock``).

    Both positional and keyword arguments (``name=value``) are supported,
    in any order.

    By default, the end tag is prefixed with "end", but that can be
    changed by passing ``end_prefix="end_"`` or similar to ``@blocktag``.

    blocktag will call the wrapped function with ``context`` and ``nodelist``
    parameters, as well as any parameters passed to the tag. It will
    also ensure that a proper error is raised if too many or too few
    parameters are passed.

    Version Changed:
        6.0:
        * Keyword arguments are now supported using ``name=value`` format.
        * The decorated function now supports keyword-only arguments.
        * Template filters are now supported for arguments.
        * Added the ``name`` argument.
        * Arbitrary arguments are deprecated and support will be removed
          in Djblets 7.

    Args:
        tag_func (callable, optional):
            The function being decorated.

            This is only set if no other arguments are provided.

        *args (tuple, unused):
            Extra positional arguments.

            This is ignored.

            Deprecated:
                7.1:
                This will be removed in Djblets 7.

        end_prefix (str, optional):
            The prefix for the end tag. This defaults to ``'end'``, but
            template tags using underscores in the name might want to change
            this to ``'end_'``.

        name (str, optional):
            An explicit name for the template tag.

            If not provided, the function name will be used.

            Version Added:
                6.0

        resolve_vars (bool, optional):
            Whether to automatically resolve all variables provided to the
            template tag. By default, variables are resolved. Template tags
            can turn this off if they want to handle variable parsing
            manually.

        **kwargs (dict, unused):
            Extra keyword arguments.

            This is ignored.

            Deprecated:
                7.1:
                This will be removed in Djblets 7.

    Returns:
        callable:
        The resulting template tag function.

    Example:
        .. code-block:: python

            @register.tag
            @blocktag
            def divify(context, nodelist, div_id=None, *,
                       css_class='my-class'):
                s = format_html('<div class="{}"', my_class)

                if div_id:
                    s += format_html(' id="{}"', div_id)

                return s + '>' + nodelist.render(context) + '</div>'
    """
    def _blocktag_func(
        tag_func: _TagFunc[_P, _R],
    ) -> _TagCompiler:
        (
            params,
            varargs,
            varkw,
            defaults,
            kwonly,
            kwonly_defaults,
            *_unused,
        ) = inspect.getfullargspec(inspect.unwrap(tag_func))

        tag_name = name or tag_func.__name__
        end_tag_name = f'{end_prefix}{tag_name}'

        @wraps(tag_func)
        def _setup_tag(
            parser: Parser,
            token: Token,
        ) -> _BlockTagNode:
            bits = token.split_contents()
            tag_args, tag_kwargs = parse_bits(
                parser,
                bits[1:],    # Skip the tag name.
                params[2:],  # Skip the context and nodelist params.
                varargs,
                varkw,
                defaults,
                kwonly,
                kwonly_defaults,
                takes_context=False,
                name=tag_name,
            )

            # Parse the contents of the block into a list of nodes.
            nodelist = parser.parse(((end_tag_name),))
            parser.delete_first_token()

            # Return our node for handling this list.
            return _BlockTagNode(
                tag_func=tag_func,
                nodelist=nodelist,
                tag_args=tag_args,
                tag_kwargs=tag_kwargs,
                resolve_vars=resolve_vars,
            )

        update_wrapper(_setup_tag, tag_func)

        return _setup_tag

    if args:
        RemovedInDjblets80Warning.warn(
            '@blocktag no longer takes extra positional arguments. This '
            'will be removed in Djblets 8.'
        )

    if kwargs:
        RemovedInDjblets80Warning.warn(
            '@blocktag no longer takes extra keyword arguments. This '
            'will be removed in Djblets 8.'
        )

    if func is not None:
        # This is being called in the @blocktag form.
        return _blocktag_func(func)
    else:
        # This is being called in the @blocktag(...) form.
        return _blocktag_func


class cached_property(django_cached_property):
    """Decorator for creating a read-only property that caches a value.

    This is a drop-in replacement for Django's
    :py:class:`~django.utils.functional.cached_property` that retains the
    docstring and attributes of the original method.

    While Django 1.8+ does retain the docstring, it does not retain the
    attributes.
    """

    def __init__(self, func):
        """Initialize the property.

        Args:
            func (callable):
                The function that will be called when this property is
                accessed. The property will have its name, documentation,
                and other attributes.
        """
        super(cached_property, self).__init__(func)

        update_wrapper(self, func)


def optional_decorator(decorator, predicate):
    """Optionally apply a decorator given a predicate function.

    Args:
        decorator (callable):
            The decorator to conditionally apply.

        predicate (callable):
            The predicate used to determine when ``decorator`` should be used.
            The arguments provided to the decorated function will be passed to
            this function.

    Returns:
        callable:
        A decorator that, when applied to a function, will call the function
        decorated with ``decorator`` when ``predicate()`` returns ``True``.
        Otherwise it will call the original function.

    Example:
        .. code-block:: python

           from djblets.util.decorators import (optional_decorator,
                                                simple_decorator)

           def predicate(verbose)
               return verbose

           @simple_decorator
           def decorator(f):
               def decorated(verbose):
                   print('Hello from the decorator')
                   return f(verbose)

           @optional_decorator(decorator, predicate)
           def f(x):
                print('Hello from f()')

           # Prints "Hello from the decorator" and "Hello from f()"
           f(verbose=True)

           # Prints "Hello from f()"
           f(verbose=False)
    """
    def _decorator(view):
        with_decorator = decorator(view)

        @wraps(view)
        def decorated(*args, **kwargs):
            if predicate(*args, **kwargs):
                return with_decorator(*args, **kwargs)

            return view(*args, **kwargs)

        return decorated

    return _decorator



"""Utilities for capturing and inspecting database queries.

Deprecated:
    6.0:
    This has been replaced with :py:mod:`django_assert_queries.query_catcher`
    from :pypi:`django-assert-queries`.

Version Added:
    3.4
"""

from __future__ import annotations

from django_assert_queries.query_catcher import (
    CatchQueriesContext,
    ExecutedQueryInfo,
    ExecutedQueryType,
    ExecutedSubQueryInfo,
    catch_queries,
)
from housekeeping import module_deprecated

from djblets.deprecation import RemovedInDjblets80Warning


module_deprecated(
    RemovedInDjblets80Warning,
    'djblets.db.query_catcher is deprecated in favor of '
    'django_assert_queries.query_catcher, and will be removed in '
    'Djblets 8. See https://pypi.org/project/django-assert-queries/')


__all__ = [
    'CatchQueriesContext',
    'ExecutedQueryInfo',
    'ExecutedQueryType',
    'ExecutedSubQueryInfo',
    'catch_queries',
]



#!/usr/bin/env python3

import os
import subprocess
import sys

scripts_dir = os.path.abspath(os.path.dirname(__file__))

# Source root directory
root_dir = os.path.abspath(os.path.join(scripts_dir, '..', '..'))
sys.path.insert(0, root_dir)

# Script config directory
sys.path.insert(0, os.path.join(scripts_dir, 'conf'))

import django
from django.core.management import call_command


if __name__ == '__main__':
    os.chdir(root_dir)

    # Verify that we have npm.
    npm_command = 'npm'

    try:
        subprocess.check_call([npm_command, '--version'],
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        raise RuntimeError(
            f'Unable to locate {npm_command} in the path, which is needed to '
            f'compile static media.'
        )

    # Set up the Django environment.
    os.putenv('FORCE_BUILD_MEDIA', '1')
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'djblets.settings')

    django.setup()

    # Install dependencies.
    subprocess.call([npm_command, 'install'])

    # Check if we're actually building media. This internal flag is used to
    # by the package build backend to better control setup vs. building of
    # static media.
    if os.environ.get('RUN_COLLECT_STATIC') != '0':
        # Build the static media.
        #
        # This will raise a CommandError or call sys.exit(1) on failure.
        call_command('collectstatic', interactive=False, verbosity=2)



"""Unit tests for djblets.protect.locks.CacheLock.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import sys

import kgb
from django.core.cache import cache

from djblets.protect.locks import CacheLock
from djblets.testing.testcases import TestCase


# Workaround for different implementations of assertNoLogs() between Django and
# the standard library.
if sys.version_info < (3, 10):
    root_logger = 'root'
else:
    root_logger = None


class CacheLockTests(kgb.SpyAgency, TestCase):
    """Unit tests for CacheLock.

    Version Added:
        6.0
    """

    def setUp(self) -> None:
        super().setUp()

        cache.clear()

    def tearDown(self) -> None:
        super().tearDown()

        cache.clear()

    def test_acquire_with_new_lock(self) -> None:
        """Testing CacheLock.acquire with new lock"""
        lock = CacheLock(key='1D5BC5F3')

        with self.assertLogs(level=logging.DEBUG) as cm:
            self.assertTrue(lock.acquire())

        self.addCleanup(lock.release)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Acquired cache lock '
            r'"example.com:1D5BC5F3" with token "[\a-z0-9-]+" for 30 '
            r'seconds \(monotonic expiration = \d+\.\d+, '
            r'estimated timestamp = \d+\.\d+\)')

        self.assertTrue(lock.acquired)
        self.assertEqual(cache.get(lock.full_cache_key), lock.token)

    def test_acquire_with_existing_lock_and_not_blocking(self) -> None:
        """Testing CacheLock.acquire with existing lock and blocking=False"""
        existing_lock = CacheLock(key='ACC54266')

        with self.assertLogs(level=logging.DEBUG) as cm:
            self.assertTrue(existing_lock.acquire())

        self.addCleanup(existing_lock.release)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Acquired cache lock '
            r'"example.com:ACC54266" with token "[\a-z0-9-]+" for 30 '
            r'seconds \(monotonic expiration = \d+\.\d+, '
            r'estimated timestamp = \d+\.\d+\)')

        new_lock = CacheLock(key='ACC54266',
                             blocking=False)
        self.assertFalse(new_lock.blocking)

        with self.assertNoLogs(logger=root_logger, level=logging.DEBUG):
            self.assertFalse(new_lock.acquire())

        self.assertFalse(new_lock.acquired)
        self.assertEqual(cache.get(new_lock.full_cache_key),
                         existing_lock.token)

    def test_acquire_with_existing_lock_and_blocking_and_released(
        self,
    ) -> None:
        """Testing CacheLock.acquire with existing lock and blocking=True and
        lock released
        """
        existing_lock = CacheLock(key='72C7A1A3',
                                  lock_expiration_secs=1)
        new_lock = CacheLock(key='72C7A1A3')

        with self.assertLogs(level=logging.DEBUG) as cm:
            self.assertTrue(existing_lock.acquire())

        self.addCleanup(existing_lock.release)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Acquired cache lock '
            r'"example.com:72C7A1A3" with token "[\a-z0-9-]+" for 1 '
            r'seconds \(monotonic expiration = \d+\.\d+, '
            r'estimated timestamp = \d+\.\d+\)')

        with self.assertLogs(level=logging.DEBUG) as cm:
            self.assertTrue(new_lock.acquire())

        self.addCleanup(new_lock.release)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Acquired cache lock '
            r'"example.com:72C7A1A3" with token "[\a-z0-9-]+" for 30 '
            r'seconds \(monotonic expiration = \d+\.\d+, '
            r'estimated timestamp = \d+\.\d+\)')

        self.assertTrue(new_lock.acquired)
        self.assertEqual(cache.get(new_lock.full_cache_key),
                         new_lock.token)

    def test_acquire_with_existing_lock_and_blocking_and_timeout(self) -> None:
        """Testing CacheLock.acquire with existing lock and blocking=True and
        lock released
        """
        existing_lock = CacheLock(key='44A816B0')
        existing_lock.acquire()
        self.addCleanup(existing_lock.release)

        new_lock = CacheLock(key='44A816B0',
                             timeout_secs=0.05)

        message = 'Timed out waiting for lock: example.com:44A816B0'

        with self.assertLogs(level=logging.DEBUG) as cm:
            with self.assertRaisesMessage(TimeoutError, message):
                self.assertFalse(new_lock.acquire())

        self.assertEqual(
            cm.output,
            [
                f'WARNING:djblets.protect.locks:Timed out waiting for cache '
                f'lock "example.com:44A816B0" (token "{new_lock.token}") '
                f'for 0.05 seconds',
            ])

        self.assertFalse(new_lock.acquired)
        self.assertEqual(cache.get(new_lock.full_cache_key),
                         existing_lock.token)

    def test_acquire_with_blocking_false(self) -> None:
        """Testing CacheLock.acquire with blocking=False overriding
        constructor
        """
        existing_lock = CacheLock(key='DF71C7BC')
        existing_lock.acquire()
        self.addCleanup(existing_lock.release)

        new_lock = CacheLock(key='DF71C7BC',
                             blocking=True,
                             timeout_secs=10)

        with self.assertNoLogs(logger=root_logger, level=logging.DEBUG):
            self.assertFalse(new_lock.acquire(blocking=False))

        self.assertFalse(new_lock.acquired)
        self.assertEqual(cache.get(new_lock.full_cache_key),
                         existing_lock.token)

    def test_release_with_acquired(self) -> None:
        """Testing CacheLock.release with lock acquired"""
        lock = CacheLock(key='F0E07CB8')
        lock.acquire()

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.release()

        self.assertEqual(
            cm.output,
            [
                f'DEBUG:djblets.protect.locks:Released cache lock '
                f'"example.com:F0E07CB8" (token "{lock.token}")'
            ])

        self.assertFalse(lock.acquired)
        self.assertIsNone(cache.get(lock.full_cache_key))

    def test_release_with_not_acquired(self) -> None:
        """Testing CacheLock.release with lock not acquired"""
        lock = CacheLock(key='CC8E30F5')

        message = 'Cannot release a lock that was not acquired.'

        with self.assertNoLogs(logger=root_logger, level=logging.DEBUG):
            with self.assertRaisesMessage(RuntimeError, message):
                lock.release()

        self.assertFalse(lock.acquired)
        self.assertIsNone(cache.get(lock.full_cache_key))

    def test_release_with_lock_token_replaced(self) -> None:
        """Testing CacheLock.release with lock token replaced"""
        lock = CacheLock(key='A49A3C30')
        lock.acquire()

        cache.set(lock.full_cache_key, 'XXX')

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.release()

        self.assertEqual(
            cm.output,
            [
                f'DEBUG:djblets.protect.locks:Released cache lock '
                f'"example.com:A49A3C30" (token "{lock.token}")'
            ])

        self.assertFalse(lock.acquired)
        self.assertEqual(cache.get(lock.full_cache_key), 'XXX')

    def test_release_with_held_beyond_expiration(self) -> None:
        """Testing CacheLock.release with lock held beyond epiration"""
        lock = CacheLock(key='4BB55758',
                         lock_expiration_secs=1)
        lock.acquire()
        lock._lock_expires_time -= 100

        self.spy_on(cache.touch)

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.release()

        self.assertEqual(
            cm.output,
            [
                f'DEBUG:djblets.protect.locks:Released cache lock '
                f'"example.com:4BB55758" (token "{lock.token}"), which '
                f'already expired from cache',
            ])

        self.assertFalse(lock.acquired)
        self.assertSpyNotCalled(cache.touch)

        # We're simulating that a new key may be in cache at the end of
        # this.
        self.assertIsNotNone(cache.get(lock.full_cache_key))

    def test_release_with_lost_key(self) -> None:
        """Testing CacheLock.release with lock key not in cache"""
        lock = CacheLock(key='84E054E5')
        lock.acquire()

        self.spy_on(cache.touch, op=kgb.SpyOpReturn(False))

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.release()

        self.assertEqual(
            cm.output,
            [
                f'DEBUG:djblets.protect.locks:Released cache lock '
                f'"example.com:84E054E5" (token '
                f'"{lock.token}"), which already expired from cache',
            ])

        self.assertFalse(lock.acquired)

        self.assertSpyCalledWith(
            cache.touch,
            'example.com:84E054E5',
            timeout=30)

        # We're simulating that a new key may be in cache at the end of
        # this.
        self.assertIsNotNone(cache.get(lock.full_cache_key))

    def test_update_expiration(self) -> None:
        """Testing CacheLock.update_expiration"""
        self.spy_on(cache.touch)

        lock = CacheLock(key='E977344F')
        lock.acquire()
        self.addCleanup(lock.release)

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.update_expiration()

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Extended cache lock '
            r'"example.com:E977344F" \(token "[a-f0-9-]+"\) to 30 '
            r'seconds \(monotonic expiration = \d+\.\d+, estimated '
            r'timestamp = \d+\.\d+\)')

        self.assertSpyCalledWith(
            cache.touch,
            'example.com:E977344F',
            timeout=30)

    def test_update_expiration_with_custom_time(self) -> None:
        """Testing CacheLock.update_expiration with custom expiration time"""
        self.spy_on(cache.touch)

        lock = CacheLock(key='587961F7')
        lock.acquire()
        self.addCleanup(lock.release)

        with self.assertLogs(level=logging.DEBUG) as cm:
            lock.update_expiration(100)

        self.assertEqual(len(cm.output), 1)
        self.assertRegex(
            cm.output[0],
            r'DEBUG:djblets.protect.locks:Extended cache lock '
            r'"example.com:587961F7" \(token "[a-f0-9-]+"\) to 100 '
            r'seconds \(monotonic expiration = \d+\.\d+, estimated '
            r'timestamp = \d+\.\d+\)')

        self.assertSpyCalledWith(
            cache.touch,
            'example.com:587961F7',
            timeout=100)

    def test_update_expiration_with_lost_lock(self) -> None:
        """Testing CacheLock.update_expiration with lost lock"""
        self.spy_on(cache.touch, op=kgb.SpyOpReturn(False))

        lock = CacheLock(key='47714EF6')
        lock.acquire()
        self.addCleanup(lock.release)

        with self.assertLogs() as cm:
            lock.update_expiration()

        self.assertEqual(
            cm.output,
            [
                f'WARNING:djblets.protect.locks:Cache lock '
                f'"example.com:47714EF6" (token "{lock.token}") fell '
                f'out of cache. The lock is no longer valid.',
            ])

        self.assertSpyCalledWith(
            cache.touch,
            'example.com:47714EF6',
            timeout=30)

    def test_context_manager(self) -> None:
        """Testing CacheLock as context manager"""
        with CacheLock(key='6E58D95B') as lock:
            self.assertTrue(lock.acquired)
            self.assertEqual(cache.get(lock.full_cache_key), lock.token)

        self.assertIsNone(cache.get(lock.full_cache_key))

    def test_del_while_acquired(self) -> None:
        """Testing CacheLock destruction while lock acquired"""
        lock = CacheLock(key='ADAEAFA1')
        lock.acquire()

        with self.assertLogs() as cm:
            del lock

        self.assertEqual(cm.output, [
            'ERROR:djblets.protect.locks:Cache lock "example.com:ADAEAFA1" '
            'was garbage collected without being released! The caller must '
            'be careful to keep this lock around until it is released.'
        ])



"""Helpers for unit tests working with features."""

from __future__ import annotations

from typing import Dict, Union

from django.test.utils import TestContextDecorator
from typing_extensions import TypeAlias

from djblets.features.feature import Feature
from djblets.features.registry import get_features_registry


#: A type mapping feature instances or IDs to enabled flags.
#:
#: Version Added:
#:     3.3
FeatureStates: TypeAlias = Dict[Union[Feature, str], bool]


class override_feature_checks(TestContextDecorator):
    """Override multiple features for a test.

    Version Changed:
        6.0:
        Changed from a pure context manager to a class that can act either as a
        context manager or a decorator for test methods.

    Version Changed:
        1.0.13:
        ``feature_states`` now accepts a
        :py:class:`~djblets.features.feature.Feature` instance as a key.

    Example:
        .. code-block:: python

           from myproject.features import my_feature_3

           feature_states = {
               'my-feature-1': True,
               'my-feature-2': False,
               my_feature_3: True,
           }

           with override_feature_checks(feature_states):
               # Your test code here.
    """

    ######################
    # Instance variables #
    ######################

    #: The desired feature states.
    feature_states: FeatureStates

    #: The old states, to restore once finished.
    _old_state: list[dict[str, object]]

    def __init__(
        self,
        feature_states: FeatureStates,
    ) -> None:
        """Initialize the override.

        Args:
            feature_states (dict):
                A dictionary of feature IDs or instances to booleans
                (representing whether the feature is enabled).
        """
        super().__init__()

        self.feature_states = feature_states
        self._old_states = []

    def enable(self) -> None:
        """Enable the feature overrides."""
        registry = get_features_registry()
        old_state: list[dict[str, object]] = []

        for feature_or_id, enabled in self.feature_states.items():
            feature: Feature | None

            if isinstance(feature_or_id, Feature):
                feature = feature_or_id
            else:
                feature = registry.get_feature(feature_or_id)
                assert feature is not None

            old_state.append({
                'feature': feature,
                'func': feature.is_enabled,
            })

            setattr(feature, 'is_enabled',
                    lambda _is_enabled=enabled, **kwargs: _is_enabled)

        self._old_state = old_state

    def disable(self) -> None:
        """Disable the feature overrides."""
        for feature_info in self._old_state:
            setattr(feature_info['feature'], 'is_enabled',
                    feature_info['func'])

        self._old_state = []


class override_feature_check(override_feature_checks):
    """Override a feature for a test.

    Unit tests can make use of this context manager to ensure that a specific
    feature has a particular enabled/disabled state before executing any code
    dependent on that feature.

    Only the provided feature will be modified, with all other feature logic
    falling back to the default behavior for the configured feature checker.

    Version Changed:
        6.0:
        Changed from a pure context manager to a class that can act either as a
        context manager or a decorator for test methods.

    Example:
        .. code-block:: python

           from myproject.features import my_feature_2

           with override_feature_check('my-feature', enabled=False):
               # Your test code here.

           with override_feature_check(my_feature_2, enabled=True):
               # Your test code here.
    """

    def __init__(
        self,
        feature_id: Feature | str,
        enabled: bool,
    ) -> None:
        """Initialize the override.

        Args:
            feature_id (str or djblets.features.feature.Feature):
                The ID or instance of the feature to override.

            enabled (bool):
                The enabled state for the feature.
        """
        super().__init__({
            feature_id: enabled,
        })



"""Management command for setting site configuration."""

from __future__ import annotations

import copy
import difflib
import json
import sys
from typing import Any, TYPE_CHECKING

from django.core.management.base import BaseCommand, CommandError
from django.utils.translation import gettext as _

from djblets.siteconfig.models import SiteConfiguration
from djblets.util.json_utils import json_merge_patch, json_patch

if TYPE_CHECKING:
    from argparse import ArgumentParser

    from djblets.siteconfig.models import (SiteConfigurationSettings,
                                           SiteConfigurationSettingsValue)


class Command(BaseCommand):
    """Sets a setting in the site configuration.

    Settings can be set using one of the following sets of options:

    :option:`--key` / :option:`--value`:
        This enables the caller to set a new value for an existing key.
        The provided key is a dot-separated path within a dictionary of
        the key to set.

        This does not allow for creating new keys.

    :option:`--json-patch`:
        This allows for setting keys to new values using the
        `JSON Patch <https://jsonpatch.com>`_ format. This can be used to
        modify existing settings, create new settings, or delete settings
        by specifying a list of operations to perform on the settings.

        .. note::

           This can be very destructive if not used correctly. Please use
           :command:`list-siteconfig` to back up your settings first.

        Version Added:
            5.2

    :option:`--json-merge-patch`:
        This allows for setting keys to new values using the
        :rfc:`JSON Merge Patch <7396>`_ format. Like JSON Patch, this can be
        used to modify existing settings, create new settings, or delete
        settings, but is done by providing a new JSON document of settings
        that can be merged into the existing document.

        .. note::

           This can be very destructive if not used correctly. Please use
           :command:`list-siteconfig` to back up your settings first.

        Version Added:
            5.2
    """

    def add_arguments(
        self,
        parser: ArgumentParser,
    ) -> None:
        """Add arguments to the command.

        Args:
            parser (object):
                The argument parser to add to.
        """
        parser.add_argument(
            '--key',
            action='store',
            dest='key',
            help=_('The existing key to modify (dot-separated).'))

        parser.add_argument(
            '--value',
            action='store',
            dest='value',
            help=_('The value to store.'))

        parser.add_argument(
            '--json-patch',
            action='store',
            dest='json_patch',
            help=_(
                'The JSON Patch to apply to settings. See '
                'https://jsonpatch.com. Use --json-patch=- to read a patch '
                'from standard input.'
            ))

        parser.add_argument(
            '--json-merge-patch',
            action='store',
            dest='json_merge_patch',
            help=_(
                'The JSON Merge Patch to apply to settings. See '
                'https://datatracker.ietf.org/doc/html/rfc7396. Use '
                '--json-merge-patch=- to read a patch from standard input.'
            ))

        parser.add_argument(
            '--dry-run',
            action='store_true',
            dest='dry_run',
            help=_(
                "Simulate applying the patch or changing a key. The results "
                "will be shown so you can make sure the changes are "
                "correct."
            ))

        parser.add_argument(
            '-y',
            '--confirm',
            action='store_true',
            dest='confirm',
            help=_(
                "Apply a patch without asking for confirmation. Please "
                "note that this can be dangerous if you haven't tested the "
                "patch first."
            ))

    def handle(self, *args, **options) -> None:
        """Handle the set-siteconfig request.

        Args:
            *args (tuple):
                Positional arguments passed to the commad.

            **options (dict):
                Options passed to the command.
        """
        siteconfig = SiteConfiguration.objects.get_current()

        confirm = options['confirm']
        dry_run = options['dry_run']
        merge_patch_json = options['json_merge_patch']
        patch_json = options['json_patch']
        patching = merge_patch_json or patch_json
        needs_confirm = patching

        if patching:
            if merge_patch_json:
                # This is a JSON Merge Patch. Load it and apply to the
                # siteconfig settings.
                patch = self._load_patch(merge_patch_json,
                                         expected_type=dict)
                new_settings = json_merge_patch(siteconfig.settings, patch)
            elif patch_json:
                # This is a JSON Patch. Load it and apply to the siteconfig
                # settings.
                patch = self._load_patch(patch_json,
                                         expected_type=list)
                new_settings = json_patch(siteconfig.settings, patch)
            else:
                assert False, 'Not reached'
        else:
            # This is a simple key/value. Set it in place of an existing key.
            key = options['key']
            value = options['value']

            if key is None:
                raise CommandError(_('--key must be provided'))

            if value is None:
                raise CommandError(_('--value must be provided'))

            new_settings = copy.deepcopy(siteconfig.settings)
            self._set_siteconfig_value(siteconfig=siteconfig,
                                       new_settings=new_settings,
                                       key=key,
                                       value=value)

        # Show the differences between the settings.
        old_settings_json = json.dumps(siteconfig.settings,
                                       indent=2,
                                       sort_keys=True)
        settings_json = json.dumps(new_settings,
                                   indent=2,
                                   sort_keys=True)

        result = list(difflib.unified_diff(
            old_settings_json.splitlines(keepends=True),
            settings_json.splitlines(keepends=True),
            fromfile='settings',
            tofile='settings'))

        if not result:
            self.stdout.write(_(
                'No settings were changed.\n'
            ))
            return

        if dry_run:
            self.stdout.write(_(
                'The following changes would be made (but --dry-run was '
                'passed):\n'
            ))
        else:
            self.stdout.write(_('The following changes will be made:\n'))

        self.stdout.write('\n')

        for line in result:
            self.stdout.write(line)

        self.stdout.write('\n')

        if not new_settings:
            raise CommandError(_(
                'The resulting settings are empty! Cowardly refusing to save.'
            ))

        # Determine whether to apply changes or prompt for confirmation.
        confirmed = (
            not dry_run and
            (not needs_confirm or
             confirm or
             (input(_(
                 'Do you want to apply these changes? [y/N] '
             )).lower() in ('y', 'yes')))
        )

        if confirmed:
            self.stdout.write('The changes have been saved.\n')
            siteconfig.settings = new_settings
            siteconfig.save(update_fields=('settings',))
        else:
            self.stdout.write('No changes have been saved.\n')

        siteconfig.save(update_fields=('settings',))

    def _load_patch(
        self,
        patch: str,
        *,
        expected_type: type,
    ) -> Any:
        """Load and validate a patch.

        Version Changed:
            5.2

        Args:
            patch (str):
                The patch to load.

                If this is "-", then the patch will be loaded from standard
                input.

            expected_type (type):
                The expected data type to load from the patch.

        Returns:
            object:
            The loaded patch.

        Raises:
            django.core.management.base.CommandError:
                There was an error loading or validating the patch.
        """
        if patch == '-':
            # Read the patch from STDIN.
            patch = sys.stdin.read()

        patch = patch.strip()

        try:
            norm_patch = json.loads(patch)
        except Exception as e:
            raise CommandError(_('Invalid patch provided: %s') % e)

        if not norm_patch:
            raise CommandError(_('No patch was provided.'))

        if not isinstance(norm_patch, expected_type):
            raise CommandError(_('Invalid patch provided. Must be a %s.')
                               % type(norm_patch).__name__)

        return norm_patch

    def _set_siteconfig_value(
        self,
        *,
        siteconfig: SiteConfiguration,
        new_settings: SiteConfigurationSettings,
        key: str,
        value: str,
    ) -> None:
        """Set a value for an existing key in the site configuration.

        This takes a dot-separated key path for an existing key in the
        site configuration settings and gives it a new value. The value is
        parsed based on the value type that exists in settings.

        Args:
            siteconfig (djblets.siteconfig.models.SiteConfiguration):
                The current site configuration, used for accessing defaults.

            new_settings (dict):
                The settings dictionary to modify.

            key (str):
                The dot-separated path to the key to set.

            value (str):
                The serialized value to parse and set on the key.

        Raises:
            django.core.management.base.CommandError:
                There was an error finding or setting the key, or validating
                the value.
        """
        path = key.split('.')
        node: SiteConfigurationSettingsValue = new_settings
        valid_key: bool = True
        key_basename: str = ''

        for item in path[:-1]:
            if not isinstance(node, dict):
                valid_key = False
                break

            try:
                node = node[item]
            except KeyError:
                valid_key = False
                break

        if valid_key:
            assert isinstance(node, dict)

            key_basename = path[-1]

            if key_basename not in node:
                valid_key = False

        if not valid_key:
            raise CommandError(_("'%s' is not a valid settings key") % key)

        # We already know this is a dict, but satisfy the type checker.
        assert isinstance(node, dict)

        stored_value = node[key_basename]
        value_type: type[SiteConfigurationSettingsValue] = type(stored_value)
        norm_value: SiteConfigurationSettingsValue = value

        if value_type not in (str, bytes, int, bool, type(None)):
            raise CommandError(_("Cannot set %s keys") % value_type.__name__)

        try:
            if value_type is bool:
                if value not in ('1', '0', 'True', 'true', 'False', 'false'):
                    raise TypeError
                else:
                    norm_value = (value in ('1', 'True', 'true'))
            elif stored_value is None:
                # Try to guess the type from any specified defaults. Otherwise
                # just assume text.
                defaults = siteconfig.get_defaults()
                value_type = type(defaults.get(key_basename, ''))

            # Special handling for 'null' -> None. If the user really wants an
            # explicit 'null' string, allow them to pass in '\null'.
            if norm_value == 'null':
                norm_value = None
            elif value == '\\null':
                norm_value = 'null'
            else:
                norm_value = value_type(value)  # type: ignore
        except TypeError:
            raise CommandError(
                _("'%(value)s' is not a valid %(type)s") % {
                    'value': value,
                    'type': value_type.__name__,
                })

        node[key_basename] = norm_value



"""General support for rate limiting of operations.

This can be used to impose rate limits on operations or access to resources,
locked down to a user, IP, or any other criteria.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import re
import time
from dataclasses import dataclass
from typing import TYPE_CHECKING

from django.core.cache import cache
from django.utils.translation import gettext as _
from typing_extensions import Self

from djblets.cache.backend import make_cache_key

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


@dataclass
class RateLimitUsage:
    """Usage information for a rate limit key.

    This information can be used to determine if a user has any activity
    remaining before hitting a rate limit, or if they've hit the limit.

    Version Added:
        6.0
    """

    #: The number of attempts made against the rate limit key.
    count: int

    #: The total number of attempts allowed.
    limit: (int | None) = None

    #: The time remaining in seconds before the rate limit resets.
    time_left_secs: (int | None) = None

    @property
    def is_limited(self) -> bool:
        """Whether the usage has hit the rate limit.

        Type:
            bool
        """
        limit = self.limit

        return limit is not None and self.count > limit


class RateLimit:
    """A rate limit applied to an operation or access.

    This represents the data behind a parsed rate limit imposed on an
    operation or access to a resource.

    Version Added:
        6.0
    """

    #: Dictionary contains keys that represent different time periods.
    #:
    #: The time periods are represented in seconds for days, hours,
    #: minutes, and seconds.
    PERIODS = {
        's': 1,
        'm': 60,
        'h': 60 * 60,
        'd': 24 * 60 * 60,
    }

    #: Regular expression that interprets the rate string.
    RATE_RE = re.compile(r'^(\d+)/(\d*)([smhd])?$')

    ######################
    # Instance variables #
    ######################

    #: The time period in seconds for a rate limit window.
    period_secs: int

    #: The total number of failed attempts allowed.
    total_limit: int

    @classmethod
    def parse(
        cls,
        rate_str: str,
    ) -> Self:
        """Return a RateLimit parsed from the given rate string.

        Converts the given rate string into a Rate object, which contains
        the number of login attempts allowed (count) and the time period
        allotted for these attempts (seconds).

        Args:
            rate (str):
                The number of attempts allowed within a period
                of time (can be seconds, minutes, hours, or days).

        Returns:
            RateLimit:
            The parsed rate limit information.
        """
        m = cls.RATE_RE.match(rate_str)

        if m:
            count, multiplier, period = m.groups()
        else:
            raise ValueError(_('Could not parse given rate limit: %s.')
                             % rate_str)

        period_secs = cls.PERIODS[period or 's']

        if multiplier:
            period_secs *= int(multiplier)

        return cls(total_limit=int(count),
                   period_secs=period_secs)

    def __init__(
        self,
        *,
        total_limit: int,
        period_secs: int,
    ) -> None:
        """Initialize attributes for the Rate object.

        This initializes the number of failed login attempts allowed, and
        the time period for the login attempts in seconds based on the
        data returned from the :py:meth:`parse` function.

        Args:
            total_limit (int):
                The total number of failed attempts allowed.

            period_secs (int):
                The time period in seconds for a rate limit window.
        """
        self.total_limit = total_limit
        self.period_secs = period_secs

    def get_reset_timestamp(
        self,
        timestamp: int,
    ) -> int:
        """Return the timestamp until a rate limit is reset.

        This calculates the number of seconds from the given timestamp when
        a rate limit period will reset.

        Args:
            timestamp (int):
                The time in seconds used as "now" for the calculation.

        Returns:
            int:
            The timestamp at which the rate limit will reset.
        """
        period_secs = self.period_secs

        if period_secs == 1:
            # The window will reset now.
            return timestamp

        return timestamp - (timestamp % period_secs) + period_secs

    def __eq__(
        self,
        other: object,
    ) -> bool:
        """Return whether the two Rate instances are equal.

        Args:
            other (object):
                The object to compare to.

        Returns:
            bool:
            Return true if the count and seconds match.
        """
        return (isinstance(other, RateLimit) and
                self.total_limit == other.total_limit and
                self.period_secs == other.period_secs)

    def __repr__(self) -> str:
        """Return a string representation of the instance.

        Returns:
            str:
            The string representation.
        """
        return (
            f'<RateLimit(period_secs={self.period_secs!r}, '
            f'total_limit={self.total_limit!r})>'
        )


def check_rate_limit(
    *,
    rate_limit: RateLimit | str | None,
    key: str | Sequence[str],
    increment_count: bool = False,
) -> RateLimitUsage:
    """Return rate limit status for a given key.

    This fetches the rate limit status for a key, optionally incrementing the
    count towards the limit in the process.

    Version Added:
        6.0

    Args:
        rate_limit (str or RateLimit):
            The default limit for any new rate limits.

            This may be a rate limit string or a pre-parsed
            :py:class:`RateLimit`.

            If ``None``, then rate limiting will not be performed. A result
            with a count of 1 will be returned.

        key (str or list of str):
            The rate limit key associated with the rate limit.

            This key is used along with other state to form a cache key. It
            may be a string or a sequence of components for a key.

        increment_count (bool, optional):
            Whether to increment the count toward the rate limit.

    Returns:
        RateLimitUsage:
        The resulting rate limit stats for the key.
    """
    if rate_limit is None:
        # If the setting is explicitly None, don't do any rate limiting.
        return RateLimitUsage(count=1)

    if isinstance(rate_limit, str):
        rate_limit = RateLimit.parse(rate_limit)

    limit = rate_limit.total_limit
    period_secs = rate_limit.period_secs

    # Prepare cache key to add or update to cache and determine remaining time
    # period left.
    now = _get_time_int()
    window = rate_limit.get_reset_timestamp(now)
    time_left_secs = window - now

    # Build a key for the cache to track usage.
    if isinstance(key, str):
        key = [key]

    cache_key = make_cache_key([
        '_ratelimit_',
        *key,
        f'{limit}/{period_secs}',
        str(window),
    ])

    # Check and track the rate limit count.
    count: (int | None) = None

    if increment_count:
        try:
            try:
                count = cache.incr(cache_key)
            except ValueError:
                # Add this to the cache, and set expiration 1 minute beyond
                # the number of seconds remaining to help avoid race
                # conditions or clock skew.
                cache.add(cache_key, 1, timeout=time_left_secs + 60)
        except Exception as e:
            logger.exception('Failed to set rate limit cache key "%s". Rate '
                             'limit checks are currently unreliable. Is the '
                             'cache server down? Error = %s',
                             cache_key, e)

    if count is None:
        try:
            count = cache.get(cache_key, 0)
        except Exception as e:
            logger.exception('Failed to fetch rate limit cache key "%s". Rate '
                             'limit checks are currently unreliable. Is the '
                             'cache server down? Error = %s',
                             cache_key, e)
            count = 0

    assert count is not None

    if not increment_count:
        # Add one to the returned value, even if we aren't incrementing the
        # stored value. This makes it so that we're consistent in how many
        # tries per period regardless of whether we're incrementing now or
        # later.
        count += 1

    return RateLimitUsage(count=count,
                          limit=limit,
                          time_left_secs=time_left_secs)


def _get_time_int() -> int:
    """Return the current time as an integer rounded to the nearest second.

    This is available as a convenience wrapper to help with unit testing.

    Version Added:
        6.0

    Returns:
        int:
        The current time rounded to the nearest second.
    """
    return int(time.time())



"""Unit tests for djblets.features.testing."""

from __future__ import annotations

from typing import ClassVar, TYPE_CHECKING

from djblets.features import Feature, FeatureLevel, get_features_registry
from djblets.features.testing import (override_feature_check,
                                      override_feature_checks)
from djblets.testing.testcases import TestCase

if TYPE_CHECKING:
    from djblets.features.testing import FeatureStates


class _StableFeature(Feature):
    feature_id: ClassVar[str] = 'djblets.tests.stable-feature'
    level = FeatureLevel.STABLE


class _ExperimentalFeature(Feature):
    feature_id: ClassVar[str] = 'djblets.tests.experimental-feature'
    level = FeatureLevel.EXPERIMENTAL


class FeatureTestingTests(TestCase):
    """Unit tests for djblets.features.testing."""

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        get_features_registry().reset()

    def test_override_feature_check(self) -> None:
        """Testing override_feature_check"""
        stable_feature = _StableFeature()
        experimental_feature = _ExperimentalFeature()

        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

        with override_feature_check(stable_feature.feature_id, enabled=False):
            self.assertFalse(stable_feature.is_enabled())
            self.assertFalse(experimental_feature.is_enabled())

        with override_feature_check(experimental_feature.feature_id,
                                    enabled=True):
            self.assertTrue(stable_feature.is_enabled())
            self.assertTrue(experimental_feature.is_enabled())

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

    def test_override_feature_checks(self) -> None:
        """Testing override_feature_checks"""
        stable_feature = _StableFeature()
        experimental_feature = _ExperimentalFeature()

        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

        feature_states: FeatureStates = {
            stable_feature.feature_id: False,
            experimental_feature.feature_id: True,
        }

        with override_feature_checks(feature_states):
            self.assertFalse(stable_feature.is_enabled())
            self.assertTrue(experimental_feature.is_enabled())

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

    def test_override_feature_checks_as_decorator(self) -> None:
        """Testing override_feature_checks as decorator"""
        stable_feature = _StableFeature()
        experimental_feature = _ExperimentalFeature()

        @override_feature_checks({
            stable_feature.feature_id: False,
            experimental_feature.feature_id: True,
        })
        def _do_test() -> None:
            self.assertFalse(stable_feature.is_enabled())
            self.assertTrue(experimental_feature.is_enabled())

        _do_test()

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

    def test_override_feature_check_as_decorator(self) -> None:
        """Testing override_feature_check as decorator"""
        stable_feature = _StableFeature()
        experimental_feature = _ExperimentalFeature()

        @override_feature_check(stable_feature.feature_id, False)
        def _do_test_1() -> None:
            self.assertFalse(stable_feature.is_enabled())

        @override_feature_check(experimental_feature.feature_id, True)
        def _do_test_2() -> None:
            self.assertTrue(experimental_feature.is_enabled())

        _do_test_1()
        _do_test_2()

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())



"""Unit tests for djblets.http.requests.

Version Added:
    6.0
"""

from __future__ import annotations

from django.test.client import RequestFactory

from djblets.http.requests import get_http_request_ip
from djblets.testing.testcases import TestCase


class GetHTTPRequestIPTests(TestCase):
    """Unit tests for get_http_request_ip.

    Version Added:
        6.0
    """

    def test_with_http_x_real_ip(self) -> None:
        """Testing get_http_request_ip with X-Real-IP header"""
        request = RequestFactory().get('/', headers={
            'X-Real-IP': '1.2.3.4',
        })

        self.assertEqual(get_http_request_ip(request), '1.2.3.4')

    def test_with_http_x_real_ip_ipv6(self) -> None:
        """Testing get_http_request_ip with X-Real-IP header with IPv6"""
        request = RequestFactory().get('/', headers={
            'X-Real-IP': '2001:db8:85a3:8d3:1319:8a2e:370:7348',
        })

        self.assertEqual(get_http_request_ip(request),
                         '2001:db8:85a3:8d3:1319:8a2e:370:7348')

    def test_with_http_x_forwarded_for(self) -> None:
        """Testing get_http_request_ip with X-Forwarded-For header"""
        request = RequestFactory().get('/', headers={
            'X-Forwarded-For': '1.2.3.4',
        })

        self.assertEqual(get_http_request_ip(request), '1.2.3.4')

    def test_with_http_x_forwarded_for_multi(self) -> None:
        """Testing get_http_request_ip with X-Forwarded-For header with
        multiple IPs
        """
        request = RequestFactory().get('/', headers={
            'X-Forwarded-For': '1.2.3.4, 4.5.5.67',
        })

        self.assertEqual(get_http_request_ip(request), '1.2.3.4')

    def test_with_http_x_forwarded_for_ipv6(self) -> None:
        """Testing get_http_request_ip with X-Forwarded-For header with IPv6
        """
        request = RequestFactory().get('/', headers={
            'X-Forwarded-For': '2001:db8:85a3:8d3:1319:8a2e:370:7348, 1.2.3.4',
        })

        self.assertEqual(get_http_request_ip(request),
                         '2001:db8:85a3:8d3:1319:8a2e:370:7348')

    def test_with__remote_addr(self) -> None:
        """Testing get_http_request_ip with REMOTE_ADDR"""
        request = RequestFactory().get('/',
                                       REMOTE_ADDR='1.2.3.4')

        self.assertEqual(get_http_request_ip(request), '1.2.3.4')

    def test_with__remote_addr_ipv6(self) -> None:
        """Testing get_http_request_ip with REMOTE_ADDR with IPv6"""
        request = RequestFactory().get(
            '/',
            REMOTE_ADDR='2001:db8:85a3:8d3:1319:8a2e:370:7348')

        self.assertEqual(get_http_request_ip(request),
                         '2001:db8:85a3:8d3:1319:8a2e:370:7348')



"""Unit tests for djblets.util.decorators.blocktag.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.template import Context, Library, NodeList, TemplateSyntaxError
from django.template.engine import Engine

from djblets.testing.testcases import TagTest
from djblets.util.decorators import blocktag

if TYPE_CHECKING:
    from djblets.util.decorators import _TagCompiler


class BlockTagTests(TagTest):
    """Unit tests for blocktag.

    Version Added:
        6.0
    """

    def test_with_no_args(self) -> None:
        """Testing blocktag with no arguments"""
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
        ) -> str:
            rendered = nodelist.render(context)

            return f'[[{rendered}]]'

        self.assertEqual(self._render_tag(my_blocktag),
                         '[[123456]]')

    def test_with_optional_pos_args(self) -> None:
        """Testing blocktag with optional positional arguments"""
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            x: int,
            y: int,
            z: str = 'z',
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y},z={z};{inner}]]'

        self.assertEqual(self._render_tag(my_blocktag, '42 100'),
                         '[[x=42,y=100,z=z;123456]]')

    def test_with_missing_required_pos_args(self) -> None:
        """Testing blocktag with missing required positional arguments
        """
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            x: int,
            y: int,
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y};{inner}]]'

        message = (
            "'my_blocktag' did not receive value(s) for the argument(s): "
            "'y'"
        )

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            self._render_tag(my_blocktag, '42')

    def test_with_keyword_as_pos_arg(self) -> None:
        """Testing blocktag with keyword argument for positional
        argument
        """
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            x: int,
            y: int,
            z: str = 'z',
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y},z={z};{inner}]]'

        self.assertEqual(self._render_tag(my_blocktag, 'x=42 y=100'),
                         '[[x=42,y=100,z=z;123456]]')

    def test_with_keyword_only_args(self) -> None:
        """Testing blocktag with keyword-only arguments"""
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            *,
            x: int,
            y: int,
            z: str = 'z',
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y},z={z};{inner}]]'

        self.assertEqual(self._render_tag(my_blocktag, 'x=42 y=100'),
                         '[[x=42,y=100,z=z;123456]]')

    def test_with_var_value(self) -> None:
        """Testing blocktag with variable for arguments"""
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            x: int,
            y: int,
            z: str = 'z',
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y},z={z};{inner}]]'

        self.assertEqual(self._render_tag(my_blocktag, 'my_var1 y=my_var2'),
                         '[[x=123,y=456,z=z;123456]]')

    def test_with_arg_filters(self) -> None:
        """Testing blocktag with filters on arguments"""
        @blocktag
        def my_blocktag(
            context: Context,
            nodelist: NodeList,
            x: int,
            y: int,
            z: str = 'z',
        ) -> str:
            inner = nodelist.render(context)

            return f'[[x={x},y={y},z={z};{inner}]]'

        self.assertEqual(self._render_tag(my_blocktag,
                                          'my_var1|add:100 y=my_var2'),
                         '[[x=223,y=456,z=z;123456]]')

    def _render_tag(
        self,
        tag_func: _TagCompiler,
        args: str = '',
    ) -> str:
        """Render a sample blocktag for testing.

        Args:
            tag_func (callable):
                The tag function to call.

            args (str):
                The arguments to pass to the template.

        Returns:
            django.utils.safestring.SafeString:
            The rendered HTML.
        """
        library = Library()
        library.tag(tag_func)

        engine = Engine()
        engine.template_builtins.append(library)

        return (
            engine.from_string(
                '{%% my_blocktag %s %%}'
                '{{my_var1}}{{my_var2}}'
                '{%% endmy_blocktag %%}'
                % args
            )
            .render(context=Context({
                'my_var1': 123,
                'my_var2': 456,
            }))
        )



"""Unit tests for djblets.conditions.values."""

from __future__ import annotations

import re
from typing import Any, Optional, Sequence

from django import forms
from django.contrib.sites.models import Site
from django.db.models.query import QuerySet
from django.http import QueryDict
from typing_extensions import assert_type

from djblets.conditions.errors import InvalidConditionValueError
from djblets.conditions.values import (ConditionValueBooleanField,
                                       ConditionValueFormField,
                                       ConditionValueIntegerField,
                                       ConditionValueModelField,
                                       ConditionValueMultipleModelField,
                                       ConditionValueRegexField)
from djblets.testing.testcases import TestCase


class ConditionValueFormFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueFormField."""

    def test_serialize_value(self) -> None:
        """Testing ConditionValueFormField.serialize_value"""
        # ModelChoiceField uses prepare_value(), needed by serialize_value(),
        # so it's a good test.
        field = ConditionValueFormField(
            forms.ModelChoiceField(queryset=Site.objects.all()))
        assert_type(field, ConditionValueFormField[Any])

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value(site), site.pk)

    def test_deserialize_value(self) -> None:
        """Testing ConditionValueFormField.deserialize_value"""
        field = ConditionValueFormField(
            forms.ModelChoiceField(queryset=Site.objects.all()))
        assert_type(field, ConditionValueFormField[Any])

        site = Site.objects.get_current()
        self.assertEqual(field.deserialize_value(site.pk), site)

    def test_get_from_form_data(self) -> None:
        """Testing ConditionValueFormField.get_from_form_data"""
        field = ConditionValueFormField[str](forms.CharField())

        data = QueryDict('test-field=abc123')
        files = QueryDict('')

        value = field.get_from_form_data(data, files, 'test-field')

        assert_type(value, Optional[str])
        self.assertEqual(value, 'abc123')

    def test_render_html(self) -> None:
        """Testing ConditionValueFormField.render_html"""
        field = ConditionValueFormField[str](forms.CharField())

        self.assertHTMLEqual(field.render_html(),
                             '<input name="XXX" type="text" />')

    def test_field_with_callable(self) -> None:
        """Testing ConditionValueFormField.field with callable field"""
        field = ConditionValueFormField[str](lambda: forms.CharField())

        self.assertIsInstance(field.field, forms.CharField)


class ConditionValueBooleanFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueBooleanField."""

    def test_deserialize_value_with_true(self) -> None:
        """Testing ConditionValueBooleanField.deserialize_value with True"""
        value_field = ConditionValueBooleanField()

        value = value_field.deserialize_value(True)
        assert_type(value, bool)
        self.assertTrue(value)

        value = value_field.deserialize_value('true')
        assert_type(value, bool)
        self.assertTrue(value)

        value = value_field.deserialize_value('True')
        assert_type(value, bool)
        self.assertTrue(value)

        value = value_field.deserialize_value('1')
        assert_type(value, bool)
        self.assertTrue(value)

    def test_deserialize_value_with_false(self) -> None:
        """Testing ConditionValueBooleanField.deserialize_value with False"""
        value_field = ConditionValueBooleanField()

        value = value_field.deserialize_value(False)
        assert_type(value, bool)
        self.assertFalse(value)

        value = value_field.deserialize_value('false')
        assert_type(value, bool)
        self.assertFalse(value)

        value = value_field.deserialize_value('False')
        assert_type(value, bool)
        self.assertFalse(value)

        value = value_field.deserialize_value('0')
        assert_type(value, bool)
        self.assertFalse(value)

    def test_serialize_value(self) -> None:
        """Testing ConditionValueBooleanField.serialize_value"""
        value_field = ConditionValueBooleanField()

        self.assertIs(value_field.serialize_value(True), True)
        self.assertIs(value_field.serialize_value(False), False)

    def test_prepare_value_for_widget(self) -> None:
        """Testing ConditionValueBooleanField.prepare_value_for_widget"""
        value_field = ConditionValueBooleanField()

        self.assertEqual(value_field.prepare_value_for_widget(True), 'True')
        self.assertEqual(value_field.prepare_value_for_widget(False), 'False')


class ConditionValueIntegerFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueIntegerField."""

    def test_deserialize_value(self) -> None:
        """Testing ConditionValueIntegerField.deserialize_value"""
        value_field = ConditionValueIntegerField()

        value = value_field.deserialize_value(100)
        assert_type(value, int)
        self.assertEqual(value, 100)

        value = value_field.deserialize_value('100')
        assert_type(value, int)
        self.assertEqual(value, 100)

    def test_serialize_value(self) -> None:
        """Testing ConditionValueIntegerField.serialize_value"""
        value_field = ConditionValueIntegerField()

        self.assertEqual(value_field.serialize_value(100), 100)


class ConditionValueModelFieldTests(TestCase):
    """Unit tests for ConditionValueModelField."""

    def test_init_with_queryset(self) -> None:
        """Testing ConditionValueModelField initialization with QuerySet"""
        field = ConditionValueModelField[Site](queryset=Site.objects.all())

        form_field = field.field
        assert isinstance(form_field, forms.ModelChoiceField)

        self.assertIsInstance(form_field.queryset, QuerySet)

    def test_init_with_callable(self) -> None:
        """Testing ConditionValueModelField initialization with callable"""
        field = ConditionValueModelField[Site](
            queryset=lambda: Site.objects.all())

        form_field = field.field
        assert isinstance(form_field, forms.ModelChoiceField)

        self.assertIsInstance(form_field.queryset, QuerySet)

    def test_serialize_value(self) -> None:
        """Testing ConditionValueModelField.serialize_value"""
        field = ConditionValueModelField[Site](queryset=Site.objects.all())

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value(site), site.pk)

    def test_deserialize_value(self) -> None:
        """Testing ConditionValueModelField.deserialize_value"""
        field = ConditionValueModelField[Site](queryset=Site.objects.all())
        site = Site.objects.get_current()

        value = field.deserialize_value(site.pk)
        assert_type(value, Site)

        self.assertEqual(value, site)


class ConditionValueMultipleModelFieldTests(TestCase):
    """Unit tests for ConditionValueMultipleModelField."""

    def test_init_with_queryset(self) -> None:
        """Testing ConditionValueMultipleModelField initialization with
        QuerySet
        """
        field = ConditionValueMultipleModelField[Site](
            queryset=Site.objects.all())

        form_field = field.field
        assert isinstance(form_field, forms.ModelMultipleChoiceField)

        self.assertIsInstance(form_field.queryset, QuerySet)

    def test_init_with_callable(self) -> None:
        """Testing ConditionValueMultipleModelField initialization with
        callable
        """
        field = ConditionValueMultipleModelField[Site](
            queryset=lambda: Site.objects.all())

        form_field = field.field
        assert isinstance(form_field, forms.ModelMultipleChoiceField)

        self.assertIsInstance(form_field.queryset, QuerySet)

    def test_serialize_value(self) -> None:
        """Testing ConditionValueMultipleModelField.serialize_value"""
        field = ConditionValueMultipleModelField[Site](
            queryset=Site.objects.all())

        site = Site.objects.get_current()
        self.assertEqual(field.serialize_value([site]), [site.pk])

    def test_deserialize_value(self) -> None:
        """Testing ConditionValueMultipleModelField.deserialize_value"""
        field = ConditionValueMultipleModelField[Site](
            queryset=Site.objects.all())

        site = Site.objects.get_current()
        result = field.deserialize_value([site.pk])

        assert_type(result, Sequence[Site])
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], site)


class ConditionValueRegexFieldTests(TestCase):
    """Unit tests for djblets.conditions.values.ConditionValueRegexField."""

    def setUp(self) -> None:
        super(ConditionValueRegexFieldTests, self).setUp()

        self.value_field = ConditionValueRegexField()

    def test_deserialize_value(self) -> None:
        """Testing ConditionValueRegexField.deserialize_value"""
        regex_obj = self.value_field.deserialize_value('ab[^c]d+e?')

        assert_type(regex_obj, re.Pattern)
        self.assertIsInstance(regex_obj, re.Pattern)
        self.assertEqual(regex_obj.pattern, re.compile('ab[^c]d+e?').pattern)

    def test_deserialize_value_with_bad_pattern(self) -> None:
        """Testing ConditionValueRegexField.deserialize_value with bad
        pattern
        """
        with self.assertRaises(InvalidConditionValueError):
            self.value_field.deserialize_value('*')

    def test_serialize_value(self) -> None:
        """Testing ConditionValueRegexField.serialize_value"""
        regex_obj = re.compile('ab[^c]d+e?')

        self.assertEqual(
            self.value_field.serialize_value(regex_obj),
            'ab[^c]d+e?')



"""Utilities for comparing executed queries to expectations.

These are used internally for unit testing, but can also be useful when
instrumenting code.

Deprecated:
    6.0:
    This has been replaced with
    :py:mod:`django_assert_queries.query_comparator` from
    :pypi:`django-assert-queries`.

Version Added:
    3.4
"""

from __future__ import annotations

from django_assert_queries.query_comparator import (
    CompareQueriesContext,
    ExpectedQueries,
    ExpectedQuery,
    QueryMismatch,
    QueryMismatchedAttr,
    compare_queries,
)
from housekeeping import module_deprecated

from djblets.deprecation import RemovedInDjblets80Warning


module_deprecated(
    RemovedInDjblets80Warning,
    'djblets.db.query_comparator is deprecated in favor of '
    'django_assert_queries.query_comparator, and will be removed in '
    'Djblets 8. See https://pypi.org/project/django-assert-queries/')


__all__ = [
    'CompareQueriesContext',
    'ExpectedQueries',
    'ExpectedQuery',
    'QueryMismatch',
    'QueryMismatchedAttr',
    'compare_queries',
]



"""Views for extensions (used in the admin UI)."""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import render
from django.views.decorators.csrf import csrf_protect
from housekeeping import deprecate_non_keyword_only_args

from djblets.deprecation import RemovedInDjblets80Warning

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any

    from django.http import HttpRequest, HttpResponse

    from djblets.extensions.extension import Extension
    from djblets.extensions.forms import SettingsForm
    from djblets.extensions.manager import ExtensionManager


@csrf_protect
@staff_member_required
@deprecate_non_keyword_only_args(RemovedInDjblets80Warning)
def extension_list(
    request: HttpRequest,
    *,
    extension_manager: ExtensionManager,
    template_name: str = 'extensions/admin/extension_list.html',
    extra_context: (Mapping[str, Any] | None) = None,
) -> HttpResponse:
    """A view for listing available extensions.

    This will load the list of extensions and render them in the provided
    template.

    It will also handle certain actions, such as forcing a reload of available
    extensions. The actions are not considered public API, and are internal
    to the page.

    Version Changed:
        6.0:
        * Added the ``extra_context`` argument.
        * All arguments except for ``request`` must be included as
          keyword-only arguments. This will be enforced in Djblets 7.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        extension_manager (djblets.extensions.manager.ExtensionManager):
            The extension manager managing the extensions shown on the page.

        template_name (str, optional):
            The name of the template to render.

        extra_context (dict, optional):
            Extra context to provide for the template.

            Version Added:
                6.0

    Returns:
        django.http.HttpResponse:
        The HTTP response to send to the client.

    Raises:
        django.http.Http404:
            The extension does not support configuration.
    """
    if request.method == 'POST':
        if 'full-reload' in request.POST:
            extension_manager.load(full_reload=True)

        return HttpResponseRedirect('.')
    else:
        # Refresh the extension list.
        extension_manager.load()

        return render(request, template_name, extra_context)


@csrf_protect
@staff_member_required
@deprecate_non_keyword_only_args(RemovedInDjblets80Warning)
def configure_extension(
    request: HttpRequest,
    *,
    ext_class: type[Extension],
    form_class: type[SettingsForm],
    extension_manager: ExtensionManager,
    template_name: str = 'extensions/configure_extension.html',
    extra_context: (Mapping[str, Any] | None) = None,
) -> HttpResponse:
    """A view for configuring an extension.

    This can be used to provide a configuration page for an extension. The
    extension can include this in its list of URL patterns, or it can wrap
    it in another function.

    The view will take care of fetching the appropriate extension, check if
    it's configurable, render it, and handle results.

    Version Changed:
        6.0:
        * Added the ``extra_context`` argument.
        * All arguments except for ``request`` must be included as
          keyword-only arguments. This will be enforced in Djblets 7.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        ext_class (type):
            The class for the extension.

        form_class (type):
            The class for the settings form.

        extension_manager (djblets.extensions.manager.ExtensionManager):
            The extension manager that manages the extension.

        template_name (str, optional):
            The name of the template to render.

        extra_context (dict, optional):
            Extra context to provide for the template.

            Version Added:
                6.0

    Returns:
        django.http.HttpResponse:
        The HTTP response to send to the client.

    Raises:
        django.http.Http404:
            The extension does not support configuration.
    """
    extension = extension_manager.get_enabled_extension(ext_class.id)

    if not extension or not extension.is_configurable:
        raise Http404

    if request.method == 'POST':
        form = form_class(extension, request.POST, request.FILES)

        if form.is_valid():
            form.save()

            return HttpResponseRedirect(f'{request.path}?saved=1')
    else:
        form = form_class(extension)

    return render(request, template_name, {
        'extension': extension,
        'form': form,
        'saved': request.GET.get('saved', 0),
        **(extra_context or {}),
    })



"""Unit tests for djblets.db.query."""

from __future__ import annotations

from django.contrib.auth.models import Group, User
from django.contrib.sites.models import Site
from django.db.models import (ForeignKey,
                              ManyToManyField,
                              ManyToManyRel,
                              ManyToOneRel,
                              OneToOneField,
                              OneToOneRel,
                              Q)

from djblets.privacy.models import StoredConsentData
from djblets.db.query import get_object_cached_field, prefix_q
from djblets.siteconfig.models import SiteConfiguration
from djblets.testing.testcases import TestCase


class GetObjectCachedFieldTests(TestCase):
    """Unit tests for get_object_cached_field().

    Version Added:
        6.0
    """

    def test_with_select_related(self) -> None:
        """Testing get_object_cached_field with select_related() value found"""
        user1 = User.objects.create(username='test-user')
        consent_data = StoredConsentData.objects.create(user=user1)

        user2 = (
            User.objects
            .filter(username='test-user')
            .select_related('storedconsentdata')
        ).first()

        assert user2 is not None

        with self.assertNumQueries(0):
            self.assertEqual(
                get_object_cached_field(user2, 'storedconsentdata'),
                consent_data)

    def test_with_select_related_and_none(self) -> None:
        """Testing get_object_cached_field with select_related() value as None
        """
        User.objects.create(username='test-user')

        user = (
            User.objects
            .filter(username='test-user')
            .select_related('storedconsentdata')
        ).first()

        assert user is not None

        with self.assertNumQueries(0):
            self.assertIsNone(get_object_cached_field(user,
                                                      'storedconsentdata'))

    def test_with_prefetch_related_fkey(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        direct ForeignKey relation with value found
        """
        self.assertIsInstance(
            SiteConfiguration._meta.get_field('site').remote_field,
            ManyToOneRel)

        site = Site.objects.get_current()
        SiteConfiguration.objects.create(site=site)

        siteconfigs = list(
            SiteConfiguration.objects
            .prefetch_related('site')
        )

        self.assertEqual(len(siteconfigs), 1)
        siteconfig = siteconfigs[0]

        with self.assertNumQueries(0):
            self.assertEqual(
                get_object_cached_field(siteconfig, 'site'),
                site)

    def test_with_prefetch_related_fkey_reverse(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        reverse ForeignKey relation with value found
        """
        self.assertIsInstance(Site._meta.get_field('config').remote_field,
                              ForeignKey)

        siteconfig = SiteConfiguration.objects.create(
            site=Site.objects.get_current())

        sites = list(
            Site.objects
            .prefetch_related('config')
        )

        self.assertEqual(len(sites), 1)
        site = sites[0]

        with self.assertNumQueries(0):
            self.assertEqual(
                get_object_cached_field(site, 'config'),
                [siteconfig])

    def test_with_prefetch_related_m2m(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        direct ManyToManyField relation with value found
        """
        self.assertIsInstance(User._meta.get_field('groups').remote_field,
                              ManyToManyRel)

        group = Group.objects.create(name='group1')

        user1 = User.objects.create(username='test-user')
        user1.groups.add(group)

        users = list(
            User.objects
            .filter(username='test-user')
            .prefetch_related('groups')
        )

        self.assertEqual(len(users), 1)
        user2 = users[0]

        with self.assertNumQueries(0):
            self.assertEqual(get_object_cached_field(user2, 'groups'),
                             [group])

    def test_with_prefetch_related_m2m_reverse(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        reverse ManyToManyField relation with value found
        """
        self.assertIsInstance(Group._meta.get_field('user').remote_field,
                              ManyToManyField)

        group = Group.objects.create(name='group1')

        user1 = User.objects.create(username='test-user')
        user1.groups.add(group)

        groups = list(
            Group.objects
            .prefetch_related('user_set')
        )

        self.assertEqual(len(groups), 1)
        group2 = groups[0]

        with self.assertNumQueries(0):
            self.assertEqual(get_object_cached_field(group2, 'user'),
                             [user1])

    def test_with_prefetch_related_oto(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        direct OneToOneField relation with value found
        """
        self.assertIsInstance(
            StoredConsentData._meta.get_field('user').remote_field,
            OneToOneRel)

        user1 = User.objects.create(username='test-user')
        StoredConsentData.objects.create(user=user1)

        consent_datas = list(
            StoredConsentData.objects
            .prefetch_related('user')
        )

        self.assertEqual(len(consent_datas), 1)
        consent_data2 = consent_datas[0]

        with self.assertNumQueries(0):
            self.assertEqual(
                get_object_cached_field(consent_data2, 'user'),
                user1)

    def test_with_prefetch_related_oto_reverse(self) -> None:
        """Testing get_object_cached_field with prefetch_related() on
        reverse OneToOneField relation with value found
        """
        self.assertIsInstance(
            User._meta.get_field('storedconsentdata').remote_field,
            OneToOneField)

        user1 = User.objects.create(username='test-user')
        consent_data = StoredConsentData.objects.create(user=user1)

        users = list(
            User.objects
            .filter(username='test-user')
            .prefetch_related('storedconsentdata')
        )

        self.assertEqual(len(users), 1)
        user2 = users[0]

        with self.assertNumQueries(0):
            self.assertEqual(
                get_object_cached_field(user2, 'storedconsentdata'),
                consent_data)


class PrefixTests(TestCase):
    """Tests for djblets.db.query.prefix_q."""

    def test_simple(self):
        """Testing prefix_q prefixes simple expressions"""
        self.assertEqual(
            str(prefix_q('fk', Q(hello='goodbye'))),
            str(Q(fk__hello='goodbye')))

    def test_nested(self):
        """Testing prefix_q prefixes nested expressions"""
        self.assertEqual(
            str(prefix_q('fk',
                         Q(foo='foo') |
                         (Q(bar='bar') &
                          Q(baz='baz')))),
            str(Q(fk__foo='foo') |
                (Q(fk__bar='bar') &
                 Q(fk__baz='baz'))))

    def test_bytestring_result(self):
        """Testing that prefix_q generates byte strings for key names"""
        q = prefix_q('fk', Q(foo='bar'))
        self.assertEqual(len(q.children), 1)
        self.assertIs(type(q.children[0]), tuple)
        self.assertIsInstance(q.children[0][0], str)



"""Unit tests for djblets.protect.ratelimit.

Version Added:
    6.0
"""

from __future__ import annotations

import kgb
from django.core.cache import cache

from djblets.protect.ratelimit import (
    RateLimit,
    RateLimitUsage,
    _get_time_int,
    check_rate_limit,
)
from djblets.testing.testcases import TestCase


class RateLimitUsageTests(TestCase):
    """Unit tests for RateLimitUsage.

    Version Added:
        6.0
    """

    def test_is_limited_without_limit(self) -> None:
        """Testing RateLimitUsage.is_limited with limit=None"""
        usage = RateLimitUsage(count=100000000000)

        self.assertIsNone(usage.limit)
        self.assertFalse(usage.is_limited)

    def test_is_limited_with_count_lte_limit(self) -> None:
        """Testing RateLimitUsage.is_limited with count <= limit"""
        usage = RateLimitUsage(count=100,
                               limit=100)

        self.assertFalse(usage.is_limited)

    def test_is_limited_with_count_gt_limit(self) -> None:
        """Testing RateLimitUsage.is_limited with count > limit"""
        usage = RateLimitUsage(count=101,
                               limit=100)

        self.assertTrue(usage.is_limited)


class RateLimitTests(TestCase):
    """Unit tests for RateLimit.

    Version Added:
        6.0
    """

    def test_parse_with_s(self) -> None:
        """Testing RateLimit.parse with /s"""
        self.assertEqual(RateLimit.parse('100/s'),
                         RateLimit(period_secs=1,
                                   total_limit=100))

    def test_parse_with_ns(self) -> None:
        """Testing RateLimit.parse with /Ns"""
        self.assertEqual(RateLimit.parse('100/10s'),
                         RateLimit(period_secs=10,
                                   total_limit=100))

    def test_parse_with_n(self) -> None:
        """Testing RateLimit.parse with /N"""
        self.assertEqual(RateLimit.parse('100/10'),
                         RateLimit(period_secs=10,
                                   total_limit=100))

    def test_parse_with_m(self) -> None:
        """Testing RateLimit.parse with /m"""
        self.assertEqual(RateLimit.parse('100/m'),
                         RateLimit(period_secs=60,
                                   total_limit=100))

    def test_parse_with_nm(self) -> None:
        """Testing RateLimit.parse with /Nm"""
        self.assertEqual(RateLimit.parse('100/10m'),
                         RateLimit(period_secs=600,
                                   total_limit=100))

    def test_parse_with_h(self) -> None:
        """Testing RateLimit.parse with /h"""
        self.assertEqual(RateLimit.parse('100/h'),
                         RateLimit(period_secs=3600,
                                   total_limit=100))

    def test_parse_with_nh(self) -> None:
        """Testing RateLimit.parse with /Nh"""
        self.assertEqual(RateLimit.parse('100/10h'),
                         RateLimit(period_secs=36000,
                                   total_limit=100))

    def test_parse_with_d(self) -> None:
        """Testing RateLimit.parse with /d"""
        self.assertEqual(RateLimit.parse('100/d'),
                         RateLimit(period_secs=86400,
                                   total_limit=100))

    def test_parse_with_nd(self) -> None:
        """Testing RateLimit.parse with /Nd"""
        self.assertEqual(RateLimit.parse('100/10d'),
                         RateLimit(period_secs=864000,
                                   total_limit=100))

    def test_parse_with_bad_string(self) -> None:
        """Testing RateLimit.parse with bad string"""
        message = 'Could not parse given rate limit: 100/f'

        with self.assertRaisesMessage(ValueError, message):
            RateLimit.parse('100/f')

    def test_get_reset_timestamp(self) -> None:
        """Testing RateLimit.get_reset_timestamp"""
        rate_limit = RateLimit(period_secs=60,
                               total_limit=100)

        self.assertEqual(rate_limit.get_reset_timestamp(60_000),
                         60_060)

    def test_get_reset_timestamp_with_period_secs_1(self) -> None:
        """Testing RateLimit.get_reset_timestamp"""
        rate_limit = RateLimit(period_secs=1,
                               total_limit=100)

        self.assertEqual(rate_limit.get_reset_timestamp(60_000),
                         60_000)


class CheckRateLimitTests(kgb.SpyAgency, TestCase):
    """Unit tests for check_rate_limit.

    Version Added:
        6.0
    """

    def setUp(self) -> None:
        """Set up state for a unit test.

        This will clear the cache.
        """
        super().setUp()

        cache.clear()

    def tearDown(self) -> None:
        """Tear down state for a unit test.

        This will clear the cache.
        """
        super().tearDown()

        cache.clear()

    def test_with_no_rate_limit(self) -> None:
        """Testing check_rate_limit with no rate limit"""
        self.assertEqual(check_rate_limit(rate_limit=None,
                                          key='test'),
                         RateLimitUsage(count=1))

    def test_with_str_rate_limit(self) -> None:
        """Testing check_rate_limit with rate limit string"""
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1759962125))

        self.assertEqual(
            check_rate_limit(rate_limit='200/5m',
                             key='test'),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

        self.assertIsNone(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'))

        # Do it one more time, to verify counts aren't retained given
        # lack of cache updates.
        self.assertEqual(
            check_rate_limit(rate_limit='200/5m',
                             key='test'),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

    def test_with_parsed_rate_limit(self) -> None:
        """Testing check_rate_limit with parsed rate limit"""
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1759962125))

        rate_limit = RateLimit(period_secs=300,
                               total_limit=200)

        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test'),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

        self.assertIsNone(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'))

        # Do it one more time, to verify counts aren't retained given
        # lack of cache updates.
        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test'),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

    def test_with_rate_limit_in_cache(self) -> None:
        """Testing check_rate_limit with rate limit in cache"""
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1759962125))

        cache.set('example.com:_ratelimit_:test:200/300:1759962300', 10)

        rate_limit = RateLimit(period_secs=300,
                               total_limit=200)

        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test'),
            RateLimitUsage(count=11,
                           limit=200,
                           time_left_secs=175))

        self.assertEqual(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'),
            10)

        # Do it one more time, to verify counts aren't retained given
        # lack of cache updates.
        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test'),
            RateLimitUsage(count=11,
                           limit=200,
                           time_left_secs=175))

    def test_with_increment_count(self) -> None:
        """Testing check_rate_limit with increment_count=True"""
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1759962125))

        rate_limit = RateLimit(period_secs=300,
                               total_limit=200)

        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test',
                             increment_count=True),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

        self.assertEqual(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'),
            1)

        # Do it one more time, to verify the counts increase beyond the
        # default.
        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test',
                             increment_count=True),
            RateLimitUsage(count=2,
                           limit=200,
                           time_left_secs=175))

        self.assertEqual(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'),
            2)

    def test_with_different_keys(self) -> None:
        """Testing check_rate_limit with different keys"""
        self.spy_on(_get_time_int, op=kgb.SpyOpReturn(1759962125))

        rate_limit = RateLimit(period_secs=300,
                               total_limit=200)

        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key='test',
                             increment_count=True),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

        self.assertEqual(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'),
            1)

        # Do it one more time, to verify the counts increase beyond the
        # default.
        self.assertEqual(
            check_rate_limit(rate_limit=rate_limit,
                             key=['other', 'test'],
                             increment_count=True),
            RateLimitUsage(count=1,
                           limit=200,
                           time_left_secs=175))

        self.assertEqual(
            cache.get('example.com:_ratelimit_:test:200/300:1759962300'),
            1)
        self.assertEqual(
            cache.get('example.com:_ratelimit_:other:test:200/300:1759962300'),
            1)



"""Unit tests for djblets.util.views.

Version Added:
    6.0
"""

from __future__ import annotations

from contextlib import contextmanager
from typing import TYPE_CHECKING

import kgb
from django.conf import settings
from django.db.utils import OperationalError
from django.test import RequestFactory, override_settings

from djblets.testing.testcases import TestCase
from djblets.util.views import (
    HealthCheckStatus,
    HealthCheckView,
    logger as views_logger,
)

if TYPE_CHECKING:
    from collections.abc import Iterator


@contextmanager
def _db_cursor() -> Iterator[None]:
    """Fake cursor method for successful tests.

    Version Added:
        6.0
    """
    yield


class HealthCheckViewTests(kgb.SpyAgency, TestCase):
    """Unit tests for djblets.util.views.HealthCheckView.

    Version Added:
        6.0
    """

    #: The request factory to use.
    factory: RequestFactory

    #: The view instance to test.
    view: HealthCheckView

    @classmethod
    def setUpClass(cls) -> None:
        """Set up the test case."""
        super().setUpClass()
        cls.factory = RequestFactory()
        cls.view = HealthCheckView()

    @classmethod
    def tearDownClass(cls) -> None:
        """Tear down the test case class."""
        del cls.factory
        del cls.view

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['192.168.1.0/24'])
    def test_get_with_forbidden_ip(self) -> None:
        """Testing HealthCheckView.get with forbidden IP address"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '10.0.0.1'

        response = self.view.get(request)

        self.assertEqual(response.status_code, 403)

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['192.168.1.100'])
    def test_get_with_allowed_ip_single(self) -> None:
        """Testing HealthCheckView.get with allowed single IP address"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '192.168.1.100'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
            ]))

        response = self.view.get(request)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content.decode(), {
            'checks': {
                'cache.default': HealthCheckStatus.UP,
                'database.default': HealthCheckStatus.UP,
            },
            'errors': {},
            'status': HealthCheckStatus.UP,
        })

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['192.168.1.0/24'])
    def test_get_with_allowed_ip_network(self) -> None:
        """Testing HealthCheckView.get with allowed IP network"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '192.168.1.50'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
            ]))

        response = self.view.get(request)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content.decode(), {
            'checks': {
                'cache.default': HealthCheckStatus.UP,
                'database.default': HealthCheckStatus.UP,
            },
            'errors': {},
            'status': HealthCheckStatus.UP,
        })

    @override_settings(
        DJBLETS_HEALTHCHECK_IPS=[
            '192.168.1.0/24',
            '10.0.0.0/8',
        ],
    )
    def test_get_with_multiple_allowed_networks(self) -> None:
        """Testing HealthCheckView.get with multiple allowed networks"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '10.0.0.5'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
            ]))

        response = self.view.get(request)

        self.assertEqual(response.status_code, 200)

    @override_settings(DJBLETS_HEALTHCHECK_IPS='192.168.1.0/24')
    def test_get_with_invalid_healthcheck_ips_setting(self) -> None:
        """Testing HealthCheckView.get with invalid DJBLETS_HEALTHCHECK_IPS"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '192.168.1.50'

        with self.assertLogs(views_logger) as logs:
            response = self.view.get(request)

        self.assertEqual(response.status_code, 403)

        assert logs is not None
        self.assertEqual(logs.output, [
            'ERROR:djblets.util.views:Health check: setting '
            '"\'192.168.1.0/24\'" for healthcheck IPs must be a list, got '
            'type "<class \'str\'>".',
        ])

    @override_settings(INTERNAL_IPS=['127.0.0.1'])
    def test_get_uses_internal_ips_fallback(self) -> None:
        """Testing HealthCheckView.get uses INTERNAL_IPS as fallback"""
        # Remove DJBLETS_HEALTHCHECK_IPS if it exists.
        del settings.DJBLETS_HEALTHCHECK_IPS

        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '127.0.0.1'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
            ]))

        response = self.view.get(request)

        self.assertEqual(response.status_code, 200)

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['127.0.0.1'])
    def test_get_with_database_connection_failure(self) -> None:
        """Testing HealthCheckView.get with database connection failure"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '127.0.0.1'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpRaise(
                        OperationalError('Connection failed')),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
            ]))

        with self.assertLogs(views_logger) as logs:
            response = self.view.get(request)

        self.assertEqual(response.status_code, 503)
        self.assertJSONEqual(response.content.decode(), {
            'checks': {
                'cache.default': HealthCheckStatus.UP,
                'database.default': HealthCheckStatus.DOWN,
            },
            'errors': {
                'database.default': 'Connection failed',
            },
            'status': HealthCheckStatus.DOWN,
        })

        self.assertEqual(logs.output, [
            'ERROR:djblets.util.views:Health check: unable to connect to '
            'database "default": Connection failed',
        ])

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['127.0.0.1'])
    def test_get_with_cache_connection_failure(self) -> None:
        """Testing HealthCheckView.get with cache connection failure"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '127.0.0.1'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['default', 'memcache']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'call_original': False,
                },
                {
                    'args': ['memcache'],
                    'op': kgb.SpyOpRaise(
                        Exception('Memcached connection failed')),
                },
            ]))

        with self.assertLogs(views_logger) as logs:
            response = self.view.get(request)

        self.assertEqual(response.status_code, 503)
        self.assertJSONEqual(response.content.decode(), {
            'checks': {
                'database.default': HealthCheckStatus.UP,
                'cache.default': HealthCheckStatus.UP,
                'cache.memcache': HealthCheckStatus.DOWN,
            },
            'errors': {
                'cache.memcache': 'Memcached connection failed',
            },
            'status': HealthCheckStatus.DOWN,
        })

        assert logs is not None
        self.assertEqual(logs.output, [
            'ERROR:djblets.util.views:Health check: unable to connect to '
            'cache "memcache": Memcached connection failed',
        ])

    @override_settings(DJBLETS_HEALTHCHECK_IPS=['127.0.0.1'])
    def test_get_successful_health_check(self) -> None:
        """Testing HealthCheckView.get with successful health check"""
        request = self.factory.get('/health/')
        request.META['REMOTE_ADDR'] = '127.0.0.1'

        self.spy_on(self.view._get_database_names,
                    op=kgb.SpyOpReturn(['default', 'secondary']))
        self.spy_on(
            self.view._get_db_cursor,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ['default'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
                {
                    'args': ['secondary'],
                    'op': kgb.SpyOpReturn(_db_cursor()),
                },
            ]))

        self.spy_on(self.view._get_cache_names,
                    op=kgb.SpyOpReturn(['memcache']))
        self.spy_on(
            self.view._check_cache,
            op=kgb.SpyOpMatchInOrder([
                {
                    'args': ('memcache',),
                    'call_original': False,
                },
            ]))

        response = self.view.get(request)

        self.assertEqual(response.status_code, 200)
        self.assertJSONEqual(response.content.decode(), {
            'checks': {
                'cache.memcache': HealthCheckStatus.UP,
                'database.default': HealthCheckStatus.UP,
                'database.secondary': HealthCheckStatus.UP,
            },
            'errors': {},
            'status': HealthCheckStatus.UP,
        })



from django.contrib.auth.models import User
from django.db import models


class MyTestExtensionModel(models.Model):
    test_field = models.CharField(max_length=16)

    class Meta(object):
        app_label = 'model_tests'
        db_table = 'model_tests_testextensionmodel'


class MyTestExtensionRelatedModel(models.Model):
    """A model relating to one outside of the extension's own app.

    This is used to check that enabling an extension rebuilds the cached
    relation trees of the models its models point to.

    Version Added:
        6.1
    """

    owner = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name='test_extension_related_models')

    class Meta(object):
        app_label = 'model_tests'
        db_table = 'model_tests_testextensionrelatedmodel'



"""Template tags for working with page states.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from django.http import HttpRequest
from django.template import Library, RequestContext
from django.utils.html import format_html_join
from django.utils.safestring import mark_safe

from djblets.pagestate.state import PageState

if TYPE_CHECKING:
    from django.template import Context
    from django.utils.safestring import SafeString


register = Library()


@register.simple_tag(takes_context=True)
def page_hook_point(
    context: Context,
    point_name: str,
) -> SafeString:
    """Register a place where content can be injected into.

    This will render any content provided by :py:mod:`page state injectors
    <djblets.pagestate.injectors>` (such as any registered
    :py:class:`~djblets.extensions.hooks.TemplateHook` for extensions) or
    by manual calls to :py:meth:`PageState.inject()
    <djblets.pagestate.state.PageState.inject>`.

    Version Added:
        6.0

    Args:
        context (django.template.Context):
            The template rendering context.

        point_name (str):
            The name of the CSS bundle to render.

    Returns:
        django.utils.safetext.SafeString:
        The rendered HTML.

    Example:
        .. code-block:: html+django

           {% load djblets_pagestate %}

           {% page_hook_point "scripts" %}
    """
    request: HttpRequest | None

    if isinstance(context, RequestContext):
        request = context.request
    else:
        request = context.get('request')

    assert request is None or isinstance(request, HttpRequest)

    if request is None:
        return mark_safe('')

    page_state = PageState.for_request(request)

    return format_html_join(
        '',
        '{}',
        (
            (page_state_data,)
            for page_state_data in page_state.iter_content(
                point_name=point_name,
                request=request,
                context=context)
        ))



"""Configures pytest and Django environment setup for Djblets.

.. important::

   Do not define plugins in this file! Plugins must be in a different
   package (such as in tests/). pytest overrides importers for plugins and
   all modules descending from that module level, which will cause extension
   importers to fail, breaking unit tests.

Version Added:
    3.0
"""

from __future__ import annotations

import os
import sys

import djblets


sys.path.insert(0, os.path.join(os.path.dirname(__file__)))


def pytest_report_header(
    *args,
    **kwargs,
) -> list[str]:
    """Return information for the report header.

    This will log the version of Djblets.

    Args:
        *args (tuple):
            Unused positional arguments.

        **kwargs (dict):
            Unused keyword arguments.

    Returns:
        list of str:
        The report header entries to log.
    """
    return [
        f'Djblets: {djblets.get_version_string()}',
    ]



"""Unit tests for djblets.log.log_timed.

Version Added:
    6.0
"""

from __future__ import annotations

import logging
import time
from contextlib import contextmanager
from typing import Any, Iterator

from django.http import HttpRequest

from djblets.log import log_timed
from djblets.testing.testcases import TestCase


root_logger = logging.getLogger()
test_logger = logging.getLogger(__name__)


class LogTimedTests(TestCase):
    """Unit tests for log_timed.

    Version Added:
        6.0
    """

    def test_as_func(self) -> None:
        """Testing log_timed as function"""
        with self._check_logs():
            t = log_timed('Doing a thing')
            t.done()

    def test_as_context_manager(self) -> None:
        """Testing log_timed as context manager"""
        with self._check_logs():
            with log_timed('Doing a thing'):
                pass

    def test_as_context_manager_and_exception(self) -> None:
        """Testing log_timed as context manager with an exception"""
        with self.assertRaisesMessage(Exception, 'oh no'):
            with self._check_logs():
                with log_timed('Doing a thing'):
                    raise Exception('oh no')

    def test_with_warning(self) -> None:
        """Testing log_timed with warning"""
        with self._check_logs(result_level_name='WARNING'):
            t = log_timed('Doing a thing',
                          warning_at=0.005)
            time.sleep(0.01)
            t.done()

    def test_with_critical(self) -> None:
        """Testing log_timed with critical"""
        with self._check_logs(result_level_name='CRITICAL'):
            t = log_timed('Doing a thing',
                          warning_at=0.002,
                          critical_at=0.005)
            time.sleep(0.01)
            t.done()

    def test_with_default_level(self) -> None:
        """Testing log_timed with default_level"""
        with self._check_logs(default_level_name='INFO',
                              result_level_name='INFO'):
            t = log_timed('Doing a thing',
                          default_level=logging.INFO)
            t.done()

    def test_with_extra(self) -> None:
        """Testing log_timed with extra="""
        with self._check_logs() as cm:
            t = log_timed(
                'Doing a thing',
                extra={
                    'foo': 'bar',
                })
            t.done()

        for record in cm.records:
            self.assertEqual(record.foo, 'bar')
            self.assertIsNone(record.request)

    def test_with_extra_and_request(self) -> None:
        """Testing log_timed with extra= and request="""
        request = HttpRequest()

        with self._check_logs() as cm:
            t = log_timed(
                'Doing a thing',
                extra={
                    'foo': 'bar',
                },
                request=request)
            t.done()

        for record in cm.records:
            self.assertEqual(record.foo, 'bar')
            self.assertIs(record.request, request)

    def test_with_logger(self) -> None:
        """Testing log_timed with logger"""
        with self._check_logs(logger_name=__name__):
            t = log_timed('Doing a thing',
                          logger=test_logger)
            t.done()

    def test_with_request(self) -> None:
        """Testing log_timed with request"""
        request = HttpRequest()

        with self._check_logs() as cm:
            t = log_timed('Doing a thing',
                          request=request)
            t.done()

        for record in cm.records:
            self.assertIs(record.request, request)

    @contextmanager
    def _check_logs(
        self,
        *,
        logger_name: str = 'root',
        default_level_name: str = 'DEBUG',
        result_level_name: str = 'DEBUG',
    ) -> Iterator[Any]:
        """Capture timed logs and run tests.

        This will capture any logs for the unit test and then check that the
        log output meets the provided criteria.

        Args:
            logger_name (str, optional):
                The name of the logger to expect in results.

            default_level_name (str, optional):
                The name of the level to expect in the begin/end log lines.

            result_level_name (str, optional):
                The name of the level to expect in the timing result log line.

        Context:
            object:
            The log assertion context for further checks.
        """
        with self.assertLogs(level=logging.DEBUG) as cm:
            yield cm

        assert cm.records[2].args

        trace_id, seconds, microseconds = cm.records[2].args

        self.assertIs(trace_id, cm.records[2].trace_id)
        self.assertIsInstance(seconds, int)
        self.assertIsInstance(microseconds, int)

        self.assertEqual(cm.output, [
            f'{default_level_name}:{logger_name}:[{trace_id}] Begin: '
            f'Doing a thing',

            f'{default_level_name}:{logger_name}:[{trace_id}] End: Doing a '
            f'thing',

            f'{result_level_name}:{logger_name}:[{trace_id}] Doing a thing '
            f'took {seconds}.{microseconds:06d} seconds',
        ])



#!/usr/bin/env python3
"""Write node.js dependencies to djblets/dependencies.py.

Version Added:
    4.0
"""

from __future__ import annotations

import itertools
import json
import os
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import TextIO


MARKER_START = '# Auto-generated Node.js dependencies {\n'
MARKER_END = '# } Auto-generated Node.js dependencies\n'


LINT_DEP_NAMES = {
    'eslint',
    '@beanbag/eslint-plugin',
}


BUILDKIT_DEP_NAMES = {
    '@beanbag/frontend-buildkit',
    '@beanbag/js-buildkit',
}


def _write_deps(
    *,
    fp: TextIO,
    doc: str,
    name: str,
    deps: Mapping[str, str],
) -> None:
    """Write dependencies to the file.

    This will write the Python code to list each dependency and the
    matching version.

    Args:
        fp (io.TextIO):
            The file to write to.

        doc (str):
            The doc comment contents.

        name (str):
            The name of the variable to write to.

        deps (dict):
            The dependencies to write.
    """
    dependencies = '\n'.join(
        f"    '{dep_name}': '{dep_ver}',"
        for dep_name, dep_ver in deps.items()

    )

    fp.write(
        f'#: {doc}\n'
        f'{name}: Mapping[str, str] = {{\n'
        f'{dependencies}\n'
        f'}}\n'
        f'\n'
    )


def main() -> None:
    """Embed package.json into djblets/dependencies.py."""
    scripts_dir = os.path.abspath(os.path.dirname(__file__))
    top_dir = os.path.abspath(os.path.join(scripts_dir, '..', '..'))
    djblets_dir = os.path.join(top_dir, 'djblets')
    package_json_path = os.path.join(djblets_dir, 'package.json')
    deps_py_path = os.path.join(djblets_dir, 'dependencies.py')

    # Load the dependencies and organize them.
    with open(package_json_path, mode='r', encoding='utf-8') as fp:
        package_json = json.load(fp)

    frontend_deps: dict[str, str] = {}
    lint_deps: dict[str, str] = {}
    npm_deps: dict[str, str] = {}

    for dep_name, dep_ver in sorted(
        itertools.chain(
            package_json['dependencies'].items(),
            package_json['devDependencies'].items(),
        )):
        if dep_name in LINT_DEP_NAMES:
            lint_deps[dep_name] = dep_ver
        elif dep_name in BUILDKIT_DEP_NAMES:
            frontend_deps[dep_name] = dep_ver
        else:
            npm_deps[dep_name] = dep_ver

    # Parse out the existing dependencies.py and grab everything outside the
    # markers.
    new_lines_pre: str = ''
    new_lines_post: str = ''

    with open(deps_py_path, mode='r', encoding='utf-8') as fp:
        data = fp.read()

        i = data.find(MARKER_START)
        assert i != -1

        j = data.find(MARKER_END, i)
        assert j != -1

        new_lines_pre = data[:i]
        new_lines_post = data[j + len(MARKER_END) + 1:]

    # Write out the new dependencies.py.
    with open(deps_py_path, mode='w', encoding='utf-8') as fp:
        fp.write(new_lines_pre)
        fp.write(f'{MARKER_START}\n\n')

        _write_deps(
            fp=fp,
            doc='Dependencies required for static media building.',
            name='frontend_buildkit_npm_dependencies',
            deps=frontend_deps)

        _write_deps(
            fp=fp,
            doc='Dependencies required for static media linting.',
            name='lint_npm_dependencies',
            deps=lint_deps)

        _write_deps(
            fp=fp,
            doc='Node dependencies required to package/develop/test Djblets.',
            name='npm_dependencies',
            deps=npm_deps)

        fp.write(
            'npm_dependencies.update(frontend_buildkit_npm_dependencies)\n'
            'npm_dependencies.update(lint_npm_dependencies)\n'
            '\n\n'
        )
        fp.write(f'{MARKER_END}\n')
        fp.write(new_lines_post)


if __name__ == '__main__':
    main()



"""Unit tests for djblets.db.fields.comma_separated_values_field."""

from __future__ import annotations

from django.core.exceptions import ValidationError
from django.db import models

from djblets.db.fields import CommaSeparatedValuesField
from djblets.testing.testcases import TestCase, TestModelsLoaderMixin


class TestModel(models.Model):
    """Test model for the CommaSeparatedValuesField."""

    field = CommaSeparatedValuesField(
        max_length=512,
        blank=True)

    choices_field = CommaSeparatedValuesField(
        choices=[
            ('a', 'Item A'),
            ('b', 'Item B'),
            ('c', 'Item C'),
        ],
        max_length=512,
        blank=True)


class CommaSeparatedValuesFieldTests(TestModelsLoaderMixin, TestCase):
    """Tests for djblets.db.fields.comma_separated_values_field."""

    tests_app = 'djblets.db.tests'

    def test_empty_field(self) -> None:
        """Testing CommaSeparatedValuesField with empty initial value."""
        obj = TestModel()
        self.assertEqual(obj.field, [])

    def test_set_value(self) -> None:
        """Testing CommaSeparatedValuesField with setting value"""
        obj = TestModel()

        obj.field = ['a', 'b', 'c']
        obj.save()

        obj.refresh_from_db()
        self.assertEqual(obj.field, ['a', 'b', 'c'])

    def test_get_list(self) -> None:
        """Testing CommaSeparatedValuesField.get_<fieldname>_list"""
        obj = TestModel()
        obj.field = ['a', 'b', 'c', 'c']

        self.assertEqual(obj.get_field_list(), ['a', 'b', 'c', 'c'])

    def test_get_list_with_choices(self) -> None:
        """Testing CommaSeparatedValuesField.get_<fieldname>_list with
        choices
        """
        obj = TestModel()
        obj.choices_field = ['a', 'b', 'c', 'c', 'z']

        self.assertEqual(obj.get_choices_field_list(), [
            'Item A',
            'Item B',
            'Item C',
            'Item C',
            'z',
        ])

    def test_get_display(self) -> None:
        """Testing CommaSeparatedValuesField.get_<fieldname>_list"""
        obj = TestModel()
        obj.field = ['a', 'b', 'c', 'c']

        self.assertEqual(obj.get_field_display(), 'a, b, c, c')

    def test_get_display_with_choices(self) -> None:
        """Testing CommaSeparatedValuesField.get_<fieldname>_list with
        choices
        """
        obj = TestModel()
        obj.choices_field = ['a', 'b', 'c', 'c', 'z']

        self.assertEqual(obj.get_choices_field_display(),
                         'Item A, Item B, Item C, Item C, z')

    def test_validate(self) -> None:
        """Testing CommaSeparatedValuesField.validate"""
        obj = TestModel()

        obj.field = ['a', 'b', 'c', 'd']
        obj.choices_field = ['a', 'b']

        obj.full_clean()

        message = '''{'choices_field': ["Value 'z' is not a valid choice."]}'''

        with self.assertRaisesMessage(ValidationError, message):
            obj.choices_field = ['a', 'b', 'z']
            obj.full_clean()



"""Unit tests for djblets.pagestate.templatetags.

Version Added:
    6.0
"""

from __future__ import annotations

from django.template import Context, Template
from django.test import RequestFactory

from djblets.pagestate.state import PageState
from djblets.testing.testcases import TestCase


class PageHookPointTests(TestCase):
    """Testing {% page_hook_point %}.

    Version Added:
        6.0
    """

    def test_with_injections(self) -> None:
        """Testing {% page_hook_point %} with injections"""
        request = RequestFactory().get('/')
        page_state = PageState.for_request(request)
        page_state.inject('test-point', {
            'content': 'Test 1\n',
        })
        page_state.inject('test-point', {
            'content': 'Test 2\n',
        })

        result = Template(
            '{% load djblets_pagestate %}'
            '{% page_hook_point "test-point" %}'
        ).render(Context({
            'request': request,
        }))

        self.assertEqual(result, 'Test 1\nTest 2\n')

    def test_without_injections(self) -> None:
        """Testing {% page_hook_point %} without injections"""
        request = RequestFactory().get('/')

        result = Template(
            '{% load djblets_pagestate %}'
            '{% page_hook_point "test-point" %}'
        ).render(Context({
            'request': request,
        }))

        self.assertEqual(result, '')



"""Utilities for working with HTTP requests.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django.http import HttpRequest


def get_http_request_ip(
    request: HttpRequest,
) -> str:
    """Return the IP address from a client for a given HTTP request.

    This will check the following headers for a suitable IP address:

    * :mailheader:`X-Real-IP`
    * :mailheader:`X-Forwarded-For`

    If not found, it will fall back to the IP address the request originated
    from, which may be a public address or an internal address within the
    service's network.

    Version Added:
        6.0

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

    Returns:
        str:
        The IP address for the request.
    """
    try:
        return request.META['HTTP_X_REAL_IP']
    except KeyError:
        try:
            return request.META['HTTP_X_FORWARDED_FOR'].split(',')[0].strip()
        except KeyError:
            return request.META['REMOTE_ADDR']



"""Various utility views."""

from __future__ import annotations

import ipaddress
import logging
from enum import Enum
from typing import TYPE_CHECKING

from django.conf import settings
from django.core.cache import caches
from django.core.cache.backends.locmem import LocMemCache
from django.db import connections
from django.db.utils import OperationalError
from django.http import HttpRequest, HttpResponseForbidden, JsonResponse
from django.utils.translation import get_language
from django.views.generic import View
from django.views.i18n import JavaScriptCatalog
from typing_extensions import TypedDict

from djblets.cache.backend import cache_memoize, make_cache_key
from djblets.cache.serials import generate_locale_serial
from djblets.util.symbols import UNSET

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from django.db.backends.utils import CursorWrapper
    from django.http.response import HttpResponse


logger = logging.getLogger(__name__)
locale_serials: dict[str, int] = {}


def cached_javascript_catalog(
    request: HttpRequest,
    domain: str = 'djangojs',
    packages: Sequence[str] = [],
) -> HttpResponse:
    """A cached version of javascript_catalog.

    Args:
        request (django.http.HttpRequest):
            The HTTP request.

        domain (str):
            The translation domain to use when looking up strings.

        packages (list of str):
            The package names to get strings from.

    Returns:
        django.http.HttpResponse:
        The response to send back to the client.
    """
    global locale_serials

    package_str = '_'.join(packages)

    try:
        serial = locale_serials[package_str]
    except KeyError:
        serial = generate_locale_serial(packages)
        locale_serials[package_str] = serial

    return cache_memoize(
        ('jsi18n', domain, package_str, get_language(), str(serial)),
        lambda: JavaScriptCatalog.as_view(domain=domain,
                                          packages=packages)(request),
        large_data=True,
        compress_large_data=True)


class HealthCheckStatus(str, Enum):
    """A status indicator for health checks.

    Version Added:
        4.0
    """
    # TODO: this can inherit from StrEnum once we're Python 3.11+

    #: The state when a service is running and can be connected to.
    UP = 'UP'

    #: The state when a service is down or otherwise cannot be connected to.
    DOWN = 'DOWN'

    #: The state when it is unknown if a service is up or down.
    UNKNOWN = 'UNKNOWN'

    def __str__(self) -> str:
        """Return a string representation of the status.

        This exists so we can compare a string against an enum. Once we move to
        StrEnum this can be removed.

        Returns:
            str:
            The string status.
        """
        return str.__str__(self)


class HealthCheckResult(TypedDict):
    """Result structure for health checks.

    Version Added:
        4.0
    """

    #: The individual checks that were run.
    checks: dict[str, HealthCheckStatus]

    #: A mapping of the check name to the error if the service is down.
    errors: dict[str, str]

    #: The overall health status of the server.
    status: HealthCheckStatus


class HealthCheckView(View):
    """A view for health checks.

    This will check the status of connected database and cache servers, and
    report whether everything can be accessed. It will return either HTTP 200
    or 503, and the payload will be a JSON blob defined by
    :py:class:`HealthCheckResult`.

    This will only allow requests from whitelisted IP addresses. These are set
    in a Django setting named ``DJBLETS_HEALTHCHECK_IPS``, which should be a
    list of strings.

    Version Changed:
        6.0:
        Changed to allow CIDR subnets in the ``DJBLETS_HEALTHCHECK_IPS``
        setting.

    Version Added:
        4.0
    """

    ######################
    # Instance variables #
    ######################

    #: The key to use for cache tests.
    #:
    #: Version Added:
    #:     6.0
    _cache_key: str

    def __init__(
        self,
        *args,
        **kwargs,
    ) -> None:
        """Initialize the view.

        Version Added:
            6.0

        Args:
            *args (tuple):
                Positional arguments to pass through to the parent class.

            **kwargs (dict):
                Keyword arguments to pass through to the parent class.
        """
        self._cache_key = make_cache_key('djblets-healthcheck')

    def get(
        self,
        request: HttpRequest,
        *args,
        **kwargs,
    ) -> HttpResponse:
        """Perform a health check.

        This will do a health check on whether the database and cache server
        can be used. If both are accessible, this will return an HTTP 200. If
        not, this will return HTTP 500.

        Version Changed:
            6.0:
            Changed IP matching to use networks instead of single IP addresses.

        Args:
            request (django.http.HttpRequest):
                The HTTP request.

            *args (tuple, unused):
                Positional arguments, for future expansion.

            **kwargs (dict, unused):
                Keyword arguments, for future expansion.

        Returns:
            django.http.JsonResponse:
            The response to send back to the client.
        """
        allowed_ips = getattr(settings, 'DJBLETS_HEALTHCHECK_IPS',
                              settings.INTERNAL_IPS)

        if not isinstance(allowed_ips, list):
            logger.error('Health check: setting "%r" for healthcheck IPs must '
                         'be a list, got type "%r".',
                         allowed_ips, type(allowed_ips))

            return HttpResponseForbidden()

        remote_addr = request.META.get('REMOTE_ADDR')
        assert isinstance(remote_addr, str)

        networks: list[ipaddress.IPv4Network | ipaddress.IPv6Network] = []

        for network in allowed_ips:
            try:
                networks.append(ipaddress.ip_network(network))
            except ValueError:
                logger.error('Health check: invalid value in setting '
                             'DJBLETS_HEALTHCHECK_IPS: %r',
                             network)

        address = ipaddress.ip_address(remote_addr)

        if not any(address in network for network in networks):
            logger.warning('Health check: got request to healthcheck endpoint '
                           'from invalid remote address %s',
                           remote_addr)

            return HttpResponseForbidden()

        result: HealthCheckResult = {
            'checks': {},
            'errors': {},
            'status': HealthCheckStatus.UP,
        }
        status: int = 200
        success: bool = True

        for key in self._get_database_names():
            result_key = f'database.{key}'

            try:
                with self._get_db_cursor(key):
                    pass

                result['checks'][result_key] = HealthCheckStatus.UP
            except OperationalError as e:
                logger.error('Health check: unable to connect to database '
                             '"%s": %s',
                             key, e,
                             extra={'request': request})

                success = False
                result['checks'][result_key] = HealthCheckStatus.DOWN
                result['errors'][result_key] = str(e)

        for cache_name in self._get_cache_names():
            if cache_name == 'forwarded_backend':
                # This is the backing for a Djblets cache forwarding backend
                # (which will be covered under another cache backend).
                # We can skip this one.
                continue

            result_key = f'cache.{cache_name}'

            try:
                self._check_cache(cache_name)

                result['checks'][result_key] = HealthCheckStatus.UP
            except Exception as e:
                logger.error('Health check: unable to connect to cache '
                             '"%s": %s',
                             cache_name, e,
                             extra={'request': request})

                success = False
                result['checks'][result_key] = HealthCheckStatus.DOWN
                result['errors'][result_key] = str(e)

        if not success:
            status = 503
            result['status'] = HealthCheckStatus.DOWN

        return JsonResponse(result, status=status)

    def _get_database_names(self) -> Iterable[str]:
        """Return a list of configured database names.

        This method exists to make it easier to do unit tests.

        Version Added:
            6.0

        Returns:
            iterator of str:
            An iterator of the database names.
        """
        return settings.DATABASES.keys()

    def _get_db_cursor(
        self,
        name: str,
    ) -> CursorWrapper:
        """Return a database cursor.

        This method exists to make it easier to do unit tests.

        Version Added:
            6.0

        Args:
            name (str):
                The name of the database to get the cursor for.

        Returns:
            django.db.backends.utils.CursorWrapper:
            The database cursor.

        Raises:
            django.db.utils.OperationalError:
                The database could not be connected.
        """
        return connections[name].cursor()

    def _get_cache_names(self) -> Iterable[str]:
        """Return a list of configured cache names.

        This method exists to make it easier to do unit tests.

        Version Added:
            6.0

        Returns:
            iterator of str:
            An iterator of the cache names.
        """
        return settings.CACHES.keys()

    def _check_cache(
        self,
        cache_name: str,
    ) -> None:
        """Check if a given cache is available.

        Args:
            cache_name (str):
                The name of the cache to test.

        Raises:
            Exception:
                The given cache was unavailable.
        """
        # We have to do this as two non-atomic operations:
        #
        # 1. Set a key explicitly.
        # 2. Check that something was set.
        #
        # The reason is that some cache backends (pymemcache notably) has
        # mitigation against intermittent outages/bad connections. They
        # can mask outages until a certain number of failures have occurred,
        # and then re-introduce the failed servers after a period of time.
        #
        # These timeframes cause problems with standard service health check
        # behavior, which often employ a Circuit Breaker pattern, requiring a
        # certain number of failures within a certain amount of time before
        # considering a service unhealthy, and then resetting when considering
        # any successful result. This does not play well with the cache
        # backend mitigations.
        #
        # Since the cache backends return a caller-provided default when in
        # outage mitigation mode, we need to check whether the key we just
        # stored can be returned or just disappears. We don't really care
        # about the value itself, just whether we got back a default.
        #
        # Note that this can have a false-positive if the cache server happens
        # to lose our newly-set key right away, but health checks should be
        # able to handle a sporadic outage report.
        cache = caches[cache_name]

        if isinstance(cache, LocMemCache):
            # This is a local memory cache. If it doesn't work, the server is
            # *really* in trouble. We can probably filter it out, though.
            return

        # Check for a key in the cache.
        cache.set(self._cache_key, True)

        if cache.get(self._cache_key, UNSET) is UNSET:
            raise Exception('Unable to communicate with cache server')



"""Injectors for page state.

Version Added:
    6.0
"""

from __future__ import annotations

from typing import Protocol, TYPE_CHECKING

from djblets.registries.registry import OrderedRegistry

if TYPE_CHECKING:
    from collections.abc import Iterator

    from django.http import HttpRequest
    from django.template import Context

    from djblets.pagestate.state import PageStateData


class PageStateInjectorProtocol(Protocol):
    """Protocol describing page state injectors.

    Classes implementing this protocol can dynamically provide data for
    page state. They must be registered with :py:attr:`page_state_injectors`.

    Version Added:
        6.0
    """

    #: The unique ID of the injector.
    injector_id: str

    def iter_page_state_data(
        self,
        *,
        point_name: str,
        request: HttpRequest,
        context: Context,
    ) -> Iterator[PageStateData]:
        """Generate page state data.

        Implementations may yield zero or more
        :py:class:`~djblets.pagestate.state.PageStateData` for a given
        page hook point name. These will be made available in the page.

        Args:
            point_name (str):
                The page hook point name to populate.

            request (django.http.HttpRequest):
                The HTTP request from the client.

            context (django.template.Context):
                The context for the templates.

        Yields:
            djblets.pagestate.state.PageStateData:
            Data for provide for the page hook point.
        """
        ...


class PageStateInjectorRegistry(OrderedRegistry[PageStateInjectorProtocol]):
    """A registry for managing page state injectors.

    Version Added:
        6.0
    """

    lookup_attrs = ('injector_id',)


#: The main registry for managing page state injectors.
#:
#: Version Added:
#:     6.0
page_state_injectors = PageStateInjectorRegistry()



"""Settings for djblets while running unit tests."""

from __future__ import annotations

import os
from pathlib import Path

from djblets.pipeline.settings import (
    build_pipeline_settings,
    find_node_modules_dirs,
)


DEBUG = True
PRODUCTION = False
TEMPLATE_DEBUG = DEBUG

ADMINS = (
    # ('Your Name', 'your_email@domain.com'),
)

MANAGERS = ADMINS

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': 'djblets_test.db',
    }
}


# Local time zone for this installation. Choices can be found here:
# http://en.wikipedia.org/wiki/List_of_tz_zones_by_name
# although not all choices may be available on all operating systems.
# If running in a Windows environment this must be set to the same as your
# system time zone.
TIME_ZONE = 'America/Chicago'

# Language code for this installation. All choices can be found here:
# http://www.i18nguy.com/unicode/language-identifiers.html
LANGUAGE_CODE = 'en-us'

SITE_ID = 1

# If you set this to False, Django will make some optimizations so as not
# to load the internationalization machinery.
USE_I18N = True

USE_TZ = True

# Absolute path to the directory that holds media.
DJBLETS_ROOT = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', 'djblets'))
HTDOCS_ROOT = os.path.join(DJBLETS_ROOT, 'htdocs')
STATIC_ROOT = os.path.join(HTDOCS_ROOT, 'static')
MEDIA_ROOT = os.path.join(HTDOCS_ROOT, 'media')

MEDIA_URL = '/media/'

# URL that handles the media served from STATIC_ROOT. Make sure to use a
# trailing slash if there is a path component (optional in other cases).
# Examples: "http://media.lawrence.com", "http://example.com/media/"
STATIC_URL = '/'

# URL prefix for admin media -- CSS, JavaScript and images. Make sure to use a
# trailing slash.
# Examples: "http://foo.com/media/", "/media/".
ADMIN_MEDIA_PREFIX = '/media/'

# Make this unique, and don't share it with anybody.
SECRET_KEY = 'af=y9ydd51a0g#bevy0+p#(7ime@m#k)$4$9imoz*!rl97w0j0'

# List of callables that know how to import templates from various sources.
TEMPLATE_LOADERS = [
    ('django.template.loaders.cached.Loader', (
        'django.template.loaders.filesystem.Loader',
        'django.template.loaders.app_directories.Loader',
    )),
]

TEMPLATE_DIRS = [
    os.path.abspath(os.path.join(os.path.dirname(__file__), 'templates')),
]

CONTEXT_PROCESSORS = []

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': TEMPLATE_DIRS,
        'OPTIONS': {
            'debug': True,
            'context_processors': CONTEXT_PROCESSORS,
            'loaders': TEMPLATE_LOADERS,
        },
    },
]

MIDDLEWARE = [
    'django.middleware.common.CommonMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
]

ROOT_URLCONF = 'djblets.testing.urls'


# Explicitly set TEST_RUNNER to avoid incorrect compatibility check warnings
# from Django 1.7.
TEST_RUNNER = 'djblets.testing.testrunners.TestRunner'


NODE_PATH = ':'.join(
    str(path)
    for path in find_node_modules_dirs(Path(DJBLETS_ROOT).parent)
)

PIPELINE = build_pipeline_settings(
    pipeline_enabled=True,
    node_modules_path=NODE_PATH,
    static_root=STATIC_ROOT,
    use_rollup=True,
    use_terser=True,
    validate_paths=True)


INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'oauth2_provider',
    'pipeline',
]


STATICFILES_DIRS = (
    ('djblets', os.path.join(STATIC_ROOT, 'djblets')),
)

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
    'djblets.extensions.staticfiles.ExtensionFinder',
    'pipeline.finders.PipelineFinder',
)

STORAGES = {
    'default': {
        'BACKEND': 'django.core.files.storage.FileSystemStorage',
    },
    'staticfiles': {
        'BACKEND': 'djblets.pipeline.storage.PipelineStorage',
    },
}


for entry in os.listdir(DJBLETS_ROOT):
    fullpath = os.path.join(DJBLETS_ROOT, entry)

    if (os.path.isdir(fullpath) and
        os.path.exists(os.path.join(fullpath, '__init__.py'))):
        INSTALLED_APPS += [f'djblets.{entry}']


INSTALLED_APPS += ['django_evolution']


OAUTH2_PROVIDER = {
    'DEFAULT_SCOPES': 'root:read',
    'SCOPES': {},
}


INTERNAL_IPS = [
    '127.0.0.1',
    '::1',
]



"""Unit tests for djblets.pagestate.middleware.PageStateMiddleware.

Version Added:
    6.0
"""

from __future__ import annotations

from django.http import HttpResponse
from django.template import Context, Template
from django.test import RequestFactory

from djblets.pagestate.state import PageState
from djblets.pagestate.middleware import PageStateMiddleware
from djblets.testing.testcases import TestCase


class PageStateMiddlewareTests(TestCase):
    """Unit tests for PageStateMiddleware.

    Version Added:
        6.0
    """

    def test_with_etag_and_injections(self) -> None:
        """Testing PageStateMiddleware with response ETag, PageState, and
        injections
        """
        request = RequestFactory().get('/')

        page_state = PageState.for_request(request)
        page_state.inject('test', {
            'content': 'Test',
            'etag': 'test',
        })

        response = HttpResponse(Template(
            '{% load djblets_pagestate %}'
            '{% page_hook_point "test" %}'
        ).render(Context({
            'request': request,
        })))
        response['ETag'] = 'test'

        new_response = PageStateMiddleware(lambda request: response)(request)

        self.assertIs(new_response, response)
        self.assertEqual(
            response['ETag'],
            '8f286193e023cd52b89af86cc9a5588f329e67c4381ec1eb6cc9c49ced471907')

    def test_without_response_etag(self) -> None:
        """Testing PageStateMiddleware without response ETag"""
        request = RequestFactory().get('/')

        page_state = PageState.for_request(request)
        page_state.inject('test', {
            'content': 'Test',
            'etag': 'test',
        })

        response = HttpResponse(Template(
            '{% load djblets_pagestate %}'
            '{% page_hook_point "test" %}'
        ).render(Context({
            'request': request,
        })))

        new_response = PageStateMiddleware(lambda request: response)(request)

        self.assertIs(new_response, response)
        self.assertNotIn('ETag', response)

    def test_without_page_state(self) -> None:
        """Testing PageStateMiddleware with response ETag and without
        PageState
        """
        request = RequestFactory().get('/')

        response = HttpResponse(Template(
            '{% load djblets_pagestate %}'
        ).render(Context({
            'request': request,
        })))
        response['ETag'] = 'test'

        new_response = PageStateMiddleware(lambda request: response)(request)

        self.assertIs(new_response, response)
        self.assertEqual(response['ETag'], 'test')

    def test_without_injections(self) -> None:
        """Testing PageStateMiddleware with response ETag and PageState,
        without injections
        """
        request = RequestFactory().get('/')

        response = HttpResponse(Template(
            '{% load djblets_pagestate %}'
            '{% page_hook_point "test" %}'
        ).render(Context({
            'request': request,
        })))
        response['ETag'] = 'test'

        new_response = PageStateMiddleware(lambda request: response)(request)

        self.assertIs(new_response, response)
        self.assertEqual(response['ETag'], 'test')
